from .gas import GAS
from .gasx import GASX
from .gasmodels import GASPoisson, GASNormal, GASLaplace, GASt, GASSkewt, GASExponential
from .gasllm import GASLLEV
from .gasllt import GASLLT
from .gasreg import GASReg

from .scores import BetatScore, SkewtScore


from math import exp, sqrt, log, tanh
import copy
import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
from scipy import optimize
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab
import seaborn as sns
import numdifftools as nd
import pandas as pd

from .covariances import acf
from .inference import BBVI, MetropolisHastings, norm_post_sim, Normal, InverseGamma, Uniform
from .output import TablePrinter
from .tests import find_p_value
from .distributions import q_Normal
from .parameter import Parameter, Parameters
from .results import BBVIResults, MLEResults, LaplaceResults, MCMCResults

class TSM(object):
    """ TSM PARENT CLASS

    Contains general time series methods to be inherited by models.

    Parameters
    ----------
    model_type : str
        The type of model (e.g. 'ARIMA', 'GARCH')
    """

    def __init__(self,model_type):

        # Holding variables for model output
        self.model_type = model_type
        self.parameters = Parameters(self.model_type)

    def _bbvi_fit(self,posterior,optimizer='RMSProp',iterations=1000,**kwargs):
        """ Performs Black Box Variational Inference

        Parameters
        ----------
        posterior : method
            Hands bbvi_fit a posterior object

        optimizer : string
            Stochastic optimizer: one of RMSProp or ADAM.

        iterations: int
            How many iterations for BBVI

        Returns
        ----------
        BBVIResults object
        """

        # Starting parameters
        phi = self.parameters.get_parameter_starting_values()
        phi = kwargs.get('start',phi).copy() # If user supplied
        batch_size = kwargs.get('batch_size',12) # If user supplied
        p = optimize.minimize(posterior,phi,method='L-BFGS-B') # PML starting values
        start_loc = 0.8*p.x + 0.2*phi
        start_ses = None

        # Starting values for approximate distribution
        for i in range(len(self.parameters.parameter_list)):
            approx_dist = self.parameters.parameter_list[i].q
            if isinstance(approx_dist, q_Normal):
                if start_ses is None:
                    self.parameters.parameter_list[i].q.loc = start_loc[i]
                    self.parameters.parameter_list[i].q.scale = -3.0
                else:
                    self.parameters.parameter_list[i].q.loc = start_loc[i]
                    self.parameters.parameter_list[i].q.scale = start_ses[i]

        q_list = [k.q for k in self.parameters.parameter_list]
        
        bbvi_obj = BBVI(posterior,q_list,batch_size,optimizer,iterations)
        q, q_params, q_ses = bbvi_obj.run()
        self.parameters.set_parameter_values(q_params,'BBVI',np.exp(q_ses),None)

        for k in range(len(self.parameters.parameter_list)):
            self.parameters.parameter_list[k].q = q[k]

        if self.model_type in ['GAS','GASX','GASLLEV','GARCH','EGARCH','EGARCHM']:
            theta, Y, scores = self._model(q_params)
            states = None
            states_var = None
            X_names = None
        elif self.model_type in ['GASLLT']:
            theta, mu_t, Y, scores = self._model(q_params)
            states = np.array([theta, mu_t])
            states_var = None
            X_names = None
        elif self.model_type in ['LMEGARCH']:
            theta, _, Y, scores = self._model(q_params)
            states = None
            states_var = None
            X_names = None    
        elif self.model_type in ['SEGARCH','SEGARCHM']:
            theta, Y, scores, y_theta = self._model(q_params)
            states = None
            states_var = None
            X_names = None    
        elif self.model_type in ['EGARCHMReg']:
            theta, Y, scores, _ = self._model(q_params)
            states = None
            states_var = None
            X_names = None           
        elif self.model_type in ['GASReg']:
            theta, Y, scores, states = self._model(q_params)
            states_var = None
            X_names = self.X_names
        elif self.model_type in ['LLEV','LLT','DynLin']:
            Y = self.data
            scores = None
            states, states_var = self.smoothed_state(self.data,q_params)
            theta = states[0][:-1]
            X_names = None  
        elif self.model_type in ['GPNARX','GPR','GP']:
            Y = self.data*self._norm_std + self._norm_mean
            scores = None
            theta = self.expected_values(self.parameters.get_parameter_values())*self._norm_std + self._norm_mean
            X_names = None  
            states = None   
            states_var = None
        else:
            theta, Y = self._model(q_params)
            scores = None
            states = None
            states_var = None
            X_names = None

        # Change this in future
        try:
            parameter_store = self.parameters.copy()
        except:
            parameter_store = self.parameters

        return BBVIResults(data_name=self.data_name,X_names=X_names,model_name=self.model_name,
            model_type=self.model_type, parameters=parameter_store,data=Y, index=self.index,
            multivariate_model=self.multivariate_model,objective_object=posterior, 
            method='BBVI',ses=q_ses,signal=theta,scores=scores,
            param_hide=self._param_hide,max_lag=self.max_lag,states=states,states_var=states_var)

    def _laplace_fit(self,obj_type):
        """ Performs a Laplace approximation to the posterior

        Parameters
        ----------
        obj_type : method
            Whether a likelihood or a posterior

        Returns
        ----------
        None (plots posterior)
        """

        # Get Mode and Inverse Hessian information
        y = self.fit(method='PML',printer=False)

        if y.ihessian is None:
            raise Exception("No Hessian information - Laplace approximation cannot be performed")
        else:
            if self.model_type in ['GAS','GASX','GASLLEV','GARCH','EGARCH','EGARCHM']:
                theta, Y, scores = self._model(y.parameters.get_parameter_values())
                states = None
                states_var = None
                X_names = None
            elif self.model_type in ['GASLLT']:
                theta, mu_t, Y, scores = self._model(y.parameters.get_parameter_values())
                states = np.array([theta, mu_t])
                states_var = None
                X_names = None        
            elif self.model_type in ['LMEGARCH']:
                theta, _, Y, scores = self._model(y.parameters.get_parameter_values())
                states = None
                states_var = None
                X_names = None        
            elif self.model_type in ['SEGARCH','SEGARCHM']:
                theta, Y, scores, y_theta = self._model(y.parameters.get_parameter_values())
                states = None
                states_var = None
                X_names = None        
            elif self.model_type in ['EGARCHMReg']:
                theta, Y, scores, _ = self._model(y.parameters.get_parameter_values())
                states = None
                states_var = None
                X_names = None                    
            elif self.model_type in ['GASReg']:
                theta, Y, scores, states = self._model(y.parameters.get_parameter_values())
                states_var = None
                X_names = self.X_names
            elif self.model_type in ['LLEV','LLT','DynLin']:
                Y = self.data
                scores = None
                states, states_var = self.smoothed_state(self.data,y.parameters.get_parameter_values())
                theta = states[0][:-1]
                X_names = None  
            elif self.model_type in ['GPNARX','GPR','GP']:
                Y = self.data*self._norm_std + self._norm_mean
                scores = None
                theta = self.expected_values(self.parameters.get_parameter_values())*self._norm_std + self._norm_mean
                X_names = None  
                states = None   
                states_var = None
            else:
                theta, Y = self._model(y.parameters.get_parameter_values())
                scores = None
                states = None
                states_var = None
                X_names = None

            # Change this in future
            try:
                parameter_store = y.parameters.copy()
            except:
                parameter_store = y.parameters

            return LaplaceResults(data_name=self.data_name,X_names=X_names,model_name=self.model_name,
                model_type=self.model_type, parameters=parameter_store,data=Y,index=self.index,
                multivariate_model=self.multivariate_model,objective_object=obj_type, 
                method='Laplace',ihessian=y.ihessian,signal=theta,scores=scores,
                param_hide=self._param_hide,max_lag=self.max_lag,states=states,states_var=states_var)

    def _mcmc_fit(self,scale=1.0,nsims=10000,printer=True,method="M-H",cov_matrix=None,**kwargs):
        """ Performs MCMC 

        Parameters
        ----------
        scale : float
            Default starting scale

        nsims : int
            Number of simulations

        printer : Boolean
            Whether to print results or not

        method : str
            What type of MCMC

        cov_matrix: None or np.array
            Can optionally provide a covariance matrix for M-H.

        Returns
        ----------
        None (plots posteriors, stores parameters)
        """
        scale = 2.38/np.sqrt(self.param_no)
        # Get Mode and Inverse Hessian information
        y = self.fit(method='PML',printer=False)
        try:
            ses = np.power(np.abs(np.diag(y.ihessian)),0.5)
            cov_matrix = np.diag(np.diag(ses))
        except:
            pass
        if method == "M-H":
            sampler = MetropolisHastings(self.neg_logposterior,scale,nsims,y.parameters.get_parameter_values(),cov_matrix=cov_matrix,model_object=None)
            chain, mean_est, median_est, upper_95_est, lower_95_est = sampler.sample()
        else:
            raise Exception("Method not recognized!")

        for k in range(len(chain)):
            chain[k] = self.parameters.parameter_list[k].prior.transform(chain[k])
            mean_est[k] = self.parameters.parameter_list[k].prior.transform(mean_est[k])
            median_est[k] = self.parameters.parameter_list[k].prior.transform(median_est[k])
            upper_95_est[k] = self.parameters.parameter_list[k].prior.transform(upper_95_est[k])
            lower_95_est[k] = self.parameters.parameter_list[k].prior.transform(lower_95_est[k])        

        self.parameters.set_parameter_values(mean_est,'M-H',None,chain)

        if self.model_type in ['GAS','GASX','GARCH','GASLLEV','EGARCH','EGARCHM']:
            theta, Y, scores = self._model(mean_est)
            states = None
            states_var = None
            X_names = None
        elif self.model_type in ['GASLLT']:
            theta, mu_t, Y, scores = self._model(mean_est)
            states = np.array([theta, mu_t])
            states_var = None
            X_names = None    
        elif self.model_type in ['LMEGARCH']:
            theta, _, Y, scores = self._model(mean_est)
            states = None
            states_var = None
            X_names = None    
        elif self.model_type in ['SEGARCH','SEGARCHM']:
            theta, Y, scores, y_theta = self._model(mean_est)
            states = None
            states_var = None
            X_names = None    
        elif self.model_type in ['EGARCHMReg']:
            theta, Y, scores, _ = self._model(mean_est)
            states = None
            states_var = None
            X_names = None                
        elif self.model_type in ['GASReg']:
            theta, Y, scores, states = self._model(mean_est)
            states_var = None
            X_names = self.X_names
        elif self.model_type in ['LLEV','LLT','DynLin']:
            Y = self.data
            scores = None
            states, states_var = self.smoothed_state(self.data,mean_est)
            theta = states[0][:-1]
            X_names = None  
        elif self.model_type in ['GPNARX','GPR','GP']:
            Y = self.data*self._norm_std + self._norm_mean
            scores = None
            theta = self.expected_values(mean_est)*self._norm_std + self._norm_mean
            X_names = None  
            states = None   
            states_var = None
        else:
            theta, Y = self._model(mean_est)
            scores = None
            states = None
            states_var = None
            X_names = None
    
        # Change this in future
        try:
            parameter_store = self.parameters.copy()
        except:
            parameter_store = self.parameters

        return MCMCResults(data_name=self.data_name,X_names=X_names,model_name=self.model_name,
            model_type=self.model_type, parameters=parameter_store,data=Y,index=self.index,
            multivariate_model=self.multivariate_model,objective_object=self.neg_logposterior, 
            method='Metropolis Hastings',samples=chain,mean_est=mean_est,median_est=median_est,lower_95_est=lower_95_est,
            upper_95_est=upper_95_est,signal=theta,scores=scores, param_hide=self._param_hide,max_lag=self.max_lag,
            states=states,states_var=states_var)

    def _ols_fit(self):
        """ Performs OLS

        Returns
        ----------
        None (stores parameters)
        """

        # TO DO - A lot of things are VAR specific here; might need to refactor in future, or just move to VAR script

        method = 'OLS'

        res_params = self._create_B_direct().flatten()
        params = res_params.copy()
        cov = self.ols_covariance()

        # Inelegant - needs refactoring
        for i in range(self.ylen):
            for k in range(self.ylen):
                if i == k or i > k:
                    params = np.append(params,self.parameters.parameter_list[-1].prior.itransform(cov[i,k]))

        ihessian = self.estimator_cov('OLS')
        res_ses = np.power(np.abs(np.diag(ihessian)),0.5)
        ses = np.append(res_ses,np.ones([params.shape[0]-res_params.shape[0]]))
        self.parameters.set_parameter_values(params,method,ses,None)

        if self.model_type in ['GAS','GASX','GARCH','EGARCH','GASLLEV','EGARCHM']:
            theta, Y, scores = self._model(params)
            states = None
            states_var = None
            X_names = None
        elif self.model_type in ['GASLLT']:
            theta, mu_t, Y, scores = self._model(params)
            states = np.array([theta, mu_t])
            states_var = None
            X_names = None    
        elif self.model_type in ['LMEGARCH']:
            theta, _, Y, scores = self._model(params)
            states = None
            states_var = None
            X_names = None    
        elif self.model_type in ['SEGARCH','SEGARCHM']:
            theta, Y, scores, y_theta = self._model(params)
            states = None
            states_var = None
            X_names = None    
        elif self.model_type in ['EGARCHMReg']:
            theta, Y, scores, _ = self._model(params)
            states = None
            states_var = None
            X_names = None                
        elif self.model_type in ['GASReg']:
            theta, Y, scores, states = self._model(params)
            X_names = self.X_names
            states_var = None
        elif self.model_type in ['LLEV','LLT','DynLin']:
            Y = self.data
            scores = None
            states, states_var = self.smoothed_state(self.data,params)
            theta = states[0][:-1]
            X_names = None
        elif self.model_type in ['GPNARX','GPR','GP']:
            Y = self.data*self._norm_std + self._norm_mean
            scores = None
            theta = self.expected_values(params)*self._norm_std + self._norm_mean
            X_names = None  
            states = None   
            states_var = None
        else:
            theta, Y = self._model(params)
            scores = None
            states = None
            X_names = None
            states_var = None

        # Change this in future
        try:
            parameter_store = self.parameters.copy()
        except:
            parameter_store = self.parameters

        return MLEResults(data_name=self.data_name,X_names=X_names,model_name=self.model_name,
            model_type=self.model_type, parameters=parameter_store,results=None,data=Y, index=self.index,
            multivariate_model=self.multivariate_model,objective_object=self.neg_loglik, 
            method=method,ihessian=ihessian,signal=theta,scores=scores,
            param_hide=self._param_hide,max_lag=self.max_lag,states=states,states_var=states_var)

    def _optimize_fit(self,obj_type=None,**kwargs):

        if obj_type == self.neg_loglik:
            method = 'MLE'
        else:
            method = 'PML'

        # Starting parameters
        phi = self.parameters.get_parameter_starting_values()
        phi = kwargs.get('start',phi).copy() # If user supplied

        # Optimize using L-BFGS-B
        p = optimize.minimize(obj_type,phi,method='L-BFGS-B')

        # Model check
        if self.model_type in ['GAS','GASX','GARCH','EGARCH','GASLLEV','EGARCHM']:
            theta, Y, scores = self._model(p.x)
            states = None
            states_var = None
            X_names = None
        elif self.model_type in ['GASLLT']:
            theta, mu_t, Y, scores = self._model(p.x)
            states = np.array([theta, mu_t])
            states_var = None
            X_names = None
        elif self.model_type in ['LMEGARCH']:
            theta, _, Y, scores = self._model(p.x)
            states = None
            states_var = None
            X_names = None    
        elif self.model_type in ['SEGARCH','SEGARCHM']:
            theta, Y, scores, y_theta = self._model(p.x)
            states = None
            states_var = None
            X_names = None    
        elif self.model_type in ['EGARCHMReg']:
            theta, Y, scores, _ = self._model(p.x)
            states = None
            states_var = None
            X_names = None                
        elif self.model_type in ['GASReg']:
            theta, Y, scores, states = self._model(p.x)
            X_names = self.X_names
            states_var = None
        elif self.model_type in ['LLEV','LLT','DynLin']:
            Y = self.data
            scores = None
            states, states_var = self.smoothed_state(self.data,p.x)
            theta = states[0][:-1]
            X_names = None
        elif self.model_type in ['GPNARX','GPR','GP']:
            Y = self.data*self._norm_std + self._norm_mean
            scores = None
            theta = self.expected_values(self.parameters.get_parameter_values())*self._norm_std + self._norm_mean
            X_names = None  
            states = None   
            states_var = None
        else:
            theta, Y = self._model(p.x)
            scores = None
            states = None
            X_names = None
            states_var = None

        # Check that matrix is non-singular; act accordingly
        try:
            ihessian = np.linalg.inv(nd.Hessian(obj_type)(p.x))
            ses = np.power(np.abs(np.diag(ihessian)),0.5)
            self.parameters.set_parameter_values(p.x,method,ses,None)

            # Change this in future
            try:
                parameter_store = self.parameters.copy()
            except:
                parameter_store = self.parameters

            return MLEResults(data_name=self.data_name,X_names=X_names,model_name=self.model_name,
                model_type=self.model_type, parameters=parameter_store,results=p,data=Y, index=self.index,
                multivariate_model=self.multivariate_model,objective_object=obj_type, 
                method=method,ihessian=ihessian,signal=theta,scores=scores,
                param_hide=self._param_hide,max_lag=self.max_lag,states=states,states_var=states_var)
        except:
            self.parameters.set_parameter_values(p.x,method,None,None)
 
            # Change this in future
            try:
                parameter_store = self.parameters.copy()
            except:
                parameter_store = self.parameters

            return MLEResults(data_name=self.data_name,X_names=X_names,model_name=self.model_name,
                model_type=self.model_type,parameters=parameter_store,results=p,data=Y, index=self.index,
                multivariate_model=self.multivariate_model,objective_object=obj_type, 
                method=method,ihessian=None,signal=theta,scores=scores,
                param_hide=self._param_hide,max_lag=self.max_lag,states=states,states_var=states_var)

    def fit(self,method=None,**kwargs):
        """ Fits a model

        Parameters
        ----------
        method : str
            A fitting method (e.g 'MLE'). Defaults to model specific default method.

        Returns
        ----------
        None (stores fit information)
        """

        cov_matrix = kwargs.get('cov_matrix',None)
        iterations = kwargs.get('iterations',1000)
        nsims = kwargs.get('nsims',10000)
        optimizer = kwargs.get('optimizer','RMSProp')
        batch_size = kwargs.get('batch_size',12)

        if method is None:
            method = self.default_method
        elif method not in self.supported_methods:
            raise ValueError("Method not supported!")

        if method == 'MLE':
            return self._optimize_fit(self.neg_loglik,**kwargs)
        elif method == 'PML':
            return self._optimize_fit(self.neg_logposterior,**kwargs)   
        elif method == 'M-H':
            return self._mcmc_fit(nsims=nsims,method=method,cov_matrix=cov_matrix)
        elif method == "Laplace":
            return self._laplace_fit(self.neg_logposterior) 
        elif method == "BBVI":
            return self._bbvi_fit(self.neg_logposterior,optimizer=optimizer,iterations=iterations,batch_size=batch_size)
        elif method == "OLS":
            return self._ols_fit()          

    def neg_logposterior(self,beta):
        """ Returns negative log posterior

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        Negative log posterior
        """

        post = self.neg_loglik(beta)
        for k in range(0,self.param_no):
            post += -self.parameters.parameter_list[k].prior.logpdf(beta[k])
        return post

    def shift_dates(self,h):
        """ Auxiliary function for creating dates for forecasts

        Parameters
        ----------
        h : int
            How many steps to forecast

        Returns
        ----------
        A transformed date_index object
        """     

        date_index = copy.deepcopy(self.index)
        date_index = date_index[self.max_lag:len(date_index)]

        if self.is_pandas is True:

            if isinstance(date_index,pd.tseries.index.DatetimeIndex):

                # Only configured for days - need to support smaller time intervals!
                for t in range(h):
                    date_index += pd.DateOffset((date_index[len(date_index)-1] - date_index[len(date_index)-2]).days)

            elif isinstance(date_index,pd.core.index.Int64Index):

                for i in range(h):
                    new_value = date_index.values[len(date_index.values)-1] + (date_index.values[len(date_index.values)-1] - date_index.values[len(date_index.values)-2])
                    date_index = pd.Int64Index(np.append(date_index.values,new_value))

        else:

            for t in range(h):
                date_index.append(date_index[len(date_index)-1]+1)

        return date_index   

    def transform_parameters(self):
        """ Transforms parameters to actual scale by applying link function

        Returns
        ----------
        Transformed parameters 
        """     
        return self.parameters.get_parameter_values(transformed=True)

    def plot_parameters(self,indices=None,figsize=(15,5),**kwargs):
        """ Plots parameters by calling parameter object

        Returns
        ----------
        Pretty plot
        """     

        self.parameters.plot_parameters(indices=indices,figsize=figsize,**kwargs)

    def adjust_prior(self,index,prior):
        """ Adjusts priors for the parameters

        Parameters
        ----------
        index : int or list[int]
            Which parameter index/indices to be altered

        prior : Prior object
            Which prior distribution? E.g. Normal(0,1)

        Returns
        ----------
        None (changes priors in Parameters object)
        """

        self.parameters.adjust_prior(index=index,prior=prior)


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab
from matplotlib import rcParams
import seaborn as sns

from .covariances import acf
from .output import TablePrinter
from .inference import *
from .distributions import *

class Parameters(object):
    """ Parameters Class

    Holds parameter objects and contains method for parameter manipulation
    """

    def __init__(self,model_name):
        self.model_name = model_name
        self.parameter_list = []
        self.estimated = False

    def __str__(self):
        param_row = []
        parameter_names = self.get_parameter_names()
        priors = self.get_parameter_priors()
        prior_names, prior_params_names = self.get_parameter_priors_names()
        vardist_names = self.get_parameter_approx_dist_names()
        transforms = self.get_parameter_transforms_names()

        fmt = [
            ('Index','param_index',6),      
            ('Parameter','param_name',25),
            ('Prior','param_prior',10),
            ('Hyperparameters','param_hyper',25),
            ('V.I. Dist','param_vardist',10),
            ('Transform','param_transform',10)
            ]       

        for param in range(len(self.parameter_list)):
            param_row.append({'param_index': param, 'param_name': parameter_names[param],
                'param_prior': prior_names[param], 'param_hyper': prior_params_names[param],
                'param_vardist': vardist_names[param], 'param_transform': transforms[param]})
        return( TablePrinter(fmt, ul='=')(param_row) )

    def add_parameter(self,name,prior,q):
        """ Adds parameter

        Parameters
        ----------
        name : str
            Name of the parameter

        prior : Prior object
            Which prior distribution? E.g. Normal(0,1)

        q : Distribution object
            Which distribution to use for variational approximation

        Returns
        ----------
        None (changes priors in Parameters object)
        """

        self.parameter_list.append(Parameter(name,len(self.parameter_list),prior,q))

    def adjust_prior(self,index,prior):
        """ Adjusts priors for the parameters

        Parameters
        ----------
        index : int or list[int]
            Which parameter index/indices to be altered

        prior : Prior object
            Which prior distribution? E.g. Normal(0,1)

        Returns
        ----------
        None (changes priors in Parameters object)
        """

        if isinstance(index, list):
            for item in index:
                if item < 0 or item > (len(self.parameter_list)-1) or not isinstance(item, int):
                    raise ValueError("Oops - the parameter index " + str(item) + " you have entered is invalid!")
                else:
                    self.parameter_list[item].prior = prior             
        else:
            if index < 0 or index > (len(self.parameter_list)-1) or not isinstance(index, int):
                raise ValueError("Oops - the parameter index " + str(index) + " you have entered is invalid!")
            else:
                self.parameter_list[index].prior = prior

    def get_parameter_names(self):
        names = []
        for parameter in self.parameter_list:
            names.append(parameter.name)
        return names

    def get_parameter_priors(self):
        priors = []
        for parameter in self.parameter_list:
            priors.append(parameter.prior)
        return priors

    def get_parameter_priors_names(self):
        priors = self.get_parameter_priors()
        prior_names = []
        prior_params_names = []
        for prior in priors:
            if isinstance(prior, Normal):
                prior_names.append('Normal')
                prior_params_names.append('mu0: ' + str(prior.mu0) + ', sigma0: ' + str(prior.sigma0))
            elif isinstance(prior, InverseGamma):
                prior_names.append('Inverse Gamma')
                prior_params_names.append('alpha: ' + str(prior.alpha) + ', beta: ' + str(prior.beta))
            elif isinstance(prior, Uniform):
                prior_names.append('Uniform')
                prior_params_names.append('n/a (non-informative)')
            else:
                raise ValueError("Prior distribution not detected!")
        return prior_names, prior_params_names

    def get_parameter_transforms(self):
        transforms = []
        for parameter in self.parameter_list:
            transforms.append(parameter.prior.transform)
        return transforms

    def get_parameter_transforms_names(self):
        transforms = []
        for parameter in self.parameter_list:
            transforms.append(parameter.prior.transform_name)
        return transforms

    def get_parameter_starting_values(self,transformed=False):
        transforms = self.get_parameter_transforms()

        values = np.zeros(len(self.parameter_list))
        for i in range(len(self.parameter_list)):
            if transformed is True:
                values[i] = transforms[i](self.parameter_list[i].start)
            else:
                values[i] = self.parameter_list[i].start
        return values

    def get_parameter_values(self,transformed=False):
        transforms = self.get_parameter_transforms()

        if self.estimated is True:
            values = np.zeros(len(self.parameter_list))
            for i in range(len(self.parameter_list)):
                if transformed is True:
                    values[i] = transforms[i](self.parameter_list[i].value)
                else:
                    values[i] = self.parameter_list[i].value
            return values
        else:
            return ValueError("No parameters have been estimated yet")

    def get_parameter_approx_dist(self):
        dists = []
        for parameter in self.parameter_list:
            dists.append(parameter.q)
        return dists

    def get_parameter_approx_dist_names(self):
        approx_dists = self.get_parameter_approx_dist()
        q_list = []

        for approx in approx_dists:
            if isinstance(approx, q_Normal):
                q_list.append('Normal')
            elif isinstance(approx, q_InverseGamma):
                q_list.append('Inverse Gamma')
            elif isinstance(approx, q_Uniform):
                q_list.append('Uniform')
            else:
                raise Exception("Approximate distribution not detected!")
        return q_list

    def set_parameter_values(self,values,method,std=None,sample=None):
        if len(values) != len(self.parameter_list):
            raise ValueError("Length of your list is not equal to length of parameters")
        for no in range(len(self.parameter_list)):
            self.parameter_list[no].method = method
            self.parameter_list[no].value = values[no]
            if std is not None:
                self.parameter_list[no].std = std[no]
            if sample is not None:
                self.parameter_list[no].sample = sample[no]
            self.estimated = True

    def set_parameter_starting_values(self,values):
        if values.shape[0] != len(self.parameter_list):
            raise ValueError("Length of your array is not equal to number of parameters")
        for no in range(len(self.parameter_list)):
            self.parameter_list[no].start = values[no]

    def plot_parameters(self,indices=None,figsize=(15,5)):
        plt.figure(figsize=figsize) 
        for parm in range(1,len(self.parameter_list)+1):
            if indices is not None and parm-1 not in indices:
                continue
            else:
                if hasattr(self.parameter_list[parm-1], 'sample'):
                    sns.distplot(self.parameter_list[parm-1].sample, rug=False, hist=False,label=self.parameter_list[parm-1].method + ' estimate of ' + self.parameter_list[parm-1].name)

                elif hasattr(self.parameter_list[parm-1], 'value') and hasattr(self.parameter_list[parm-1], 'std'): 

                    if self.parameter_list[parm-1].prior.transform_name is None:
                        x = np.linspace(self.parameter_list[parm-1].value-self.parameter_list[parm-1].std*3.5,self.parameter_list[parm-1].value+self.parameter_list[parm-1].std*3.5,100)
                        plt.plot(x,mlab.normpdf(x,self.parameter_list[parm-1].value,self.parameter_list[parm-1].std),label=self.parameter_list[parm-1].method + ' estimate of ' + self.parameter_list[parm-1].name)
                    else:
                        sims = self.parameter_list[parm-1].prior.transform(np.random.normal(self.parameter_list[parm-1].value,self.parameter_list[parm-1].std,100000))
                        sns.distplot(sims, rug=False, hist=False,label=self.parameter_list[parm-1].method + ' estimate of ' + self.parameter_list[parm-1].name)


                else:
                    raise ValueError("No information on parameter to plot!")        

        plt.xlabel('Value')
        plt.ylabel('Frequency')
        plt.title('Parameter Plot')
        plt.legend()
        plt.show()

    def trace_plot(self,figsize=(15,15)):
        if hasattr(self.parameter_list[0], 'sample'):
            fig = plt.figure(figsize=figsize)
            
            palette = [(0.2980392156862745, 0.4470588235294118, 0.6901960784313725), 
            (0.3333333333333333, 0.6588235294117647, 0.40784313725490196), 
            (0.7686274509803922, 0.3058823529411765, 0.3215686274509804), 
            (0.5058823529411764, 0.4470588235294118, 0.6980392156862745), 
            (0.8, 0.7254901960784313, 0.4549019607843137), 
            (0.39215686274509803, 0.7098039215686275, 0.803921568627451)] * len(self.parameter_list)
            
            for j in range(len(self.parameter_list)):
                chain = self.parameter_list[j].sample
                for k in range(4):
                    iteration = j*4 + k + 1
                    ax = fig.add_subplot(len(self.parameter_list),4,iteration)
                    if iteration in range(1,len(self.parameter_list)*4 + 1,4):
                        a = sns.distplot(chain, rug=False, hist=False,color=palette[j])
                        a.set_ylabel(self.parameter_list[j].name)
                        if iteration == 1:
                            a.set_title('Density Estimate')
                    elif iteration in range(2,len(self.parameter_list)*4 + 1,4):
                        a = plt.plot(chain,color=palette[j])
                        if iteration == 2:
                            plt.title('Trace Plot')
                    elif iteration in range(3,len(self.parameter_list)*4 + 1,4): 
                        plt.plot(np.cumsum(chain)/np.array(range(1,len(chain)+1)),color=palette[j])
                        if iteration == 3:
                            plt.title('Cumulative Average')                 
                    elif iteration in range(4,len(self.parameter_list)*4 + 1,4):
                        plt.bar(range(1,10),[acf(chain,lag) for lag in range(1,10)],color=palette[j])
                        if iteration == 4:
                            plt.title('ACF Plot')                       
            sns.plt.show()  
        else:
            raise ValueError("No samples to plot!")


class Parameter(object):
    """ Parameter Class

    Parameters
    ----------
    name : str
        Name of the parameter

    index : int
        Index of the parameter

    prior : Prior object
        The prior for the parameter, e.g. Normal(0,1)

    q : dst.q_[dist] object
        The variational distribution for the parameter, e.g. q_Normal(0,1)
    """

    def __init__(self,name,index,prior,q):
        self.name = name
        self.index = 0
        self.prior = prior
        self.transform = self.prior.transform
        self.start = 0.0
        self.q = q

    def plot_parameter(self,figsize=(15,5)):
        if hasattr(self, 'sample'):
            sns.distplot(self.sample, rug=False, hist=False,label=self.method + ' estimate of ' + self.name)

        elif hasattr(self, 'value') and hasattr(self, 'std'):
            x = np.linspace(self.value-self.std*3.5,self.value+self.std*3.5,100)
            plt.figure(figsize=figsize)
            if self.prior.transform_name is None:
                plt.plot(x,mlab.normpdf(x,self.value,self.std),label=self.method + ' estimate of ' + self.name)
            else:
                sims = self.prior.transform(np.random.normal(self.value,self.std,100000))
                sns.distplot(sims, rug=False, hist=False,label=self.method + ' estimate of ' + self.name)
            plt.xlabel('Value')
            plt.legend()
            plt.show()

        else:
            raise ValueError("No information on parameter to plot!")


import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
import matplotlib.pyplot as plt
import seaborn as sns
from patsy import dmatrices, dmatrix, demo_data

from .. import inference as ifr
from .. import distributions as dst
from .. import output as op
from .. import tests as tst
from .. import tsm as tsm
from .. import data_check as dc

class ARIMAX(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** AUTOREGRESSIVE INTEGRATED MOVING AVERAGE EXOGENOUS (ARIMAX) MODELS ****

    Parameters
    ----------
    data : pd.DataFrame or np.array
        Field to specify the time series data that will be used.

    formula : string
        patsy string describing the regression

    ar : int
        Field to specify how many AR terms the model will have.

    ma : int
        Field to specify how many MA terms the model will have.

    integ : int (default : 0)
        Specifies how many times to difference the time series.
    """

    def __init__(self,data,formula,ar,ma,integ=0):

        # Initialize TSM object
        super(ARIMAX,self).__init__('ARIMAX')

        # Parameters
        self.ar = ar
        self.ma = ma
        self.integ = integ
        self.model_name = "ARIMAX(" + str(self.ar) + "," + str(self.integ) + "," + str(self.ma) + ")"
        self.param_no = self.ar + self.ma + 2
        self.max_lag = max(self.ar,self.ma)
        self._param_hide = 0 # Whether to cutoff variance parameters from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"
        self.multivariate_model = False

        # Format the data
        self.is_pandas = True # This is compulsory for this model type
        self.data_original = data.copy()
        self.formula = formula
        self.y, self.X = dmatrices(formula, data)
        self.param_no = self.X.shape[1]
        self.y_name = self.y.design_info.describe()
        self.data_name = self.y_name
        self.X_names = self.X.design_info.describe().split(" + ")
        self.y = np.array([self.y]).ravel()
        self.data = self.y.copy()
        self.X = np.array([self.X])[0]
        self.index = data.index

        # Difference data
        for order in range(0,self.integ):
            self.y = np.diff(self.y)
            self.data = np.diff(self.data)
            self.data_name = "Differenced " + self.data_name

        self.ar_matrix = self._ar_matrix()
        self._create_parameters()

    def _ar_matrix(self):
        """ Creates Autoregressive Matrix

        Returns
        ----------
        X : np.array
            Autoregressive Matrix

        """

        if self.ar != 0:
            X = self.data[(self.max_lag-1):-1]
            for i in range(1,self.ar):
                X = np.vstack((X,self.data[(self.max_lag-i-1):-i-1]))
            return X
        else:
            return np.zeros(self.data.shape[0]-self.max_lag)

    def _create_parameters(self):
        """ Creates model parameters

        Returns
        ----------
        None (changes model attributes)
        """

        for ar_term in range(self.ar):
            self.parameters.add_parameter('AR(' + str(ar_term+1) + ')',ifr.Normal(0,0.5,transform=None),dst.q_Normal(0,3))

        for ma_term in range(self.ma):
            self.parameters.add_parameter('MA(' + str(ma_term+1) + ')',ifr.Normal(0,0.5,transform=None),dst.q_Normal(0,3))

        for parm in range(len(self.X_names)):
            self.parameters.add_parameter('Beta ' + self.X_names[parm],ifr.Normal(0,3,transform=None),dst.q_Normal(0,3))

        self.parameters.add_parameter('Sigma',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))

    def _model(self,beta):
        """ Creates the structure of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        mu : np.array
            Contains the predicted values for the time series

        Y : np.array
            Contains the length-adjusted time series (accounting for lags)
        """     

        Y = self.y[self.max_lag:]

        # Transform parameters
        parm = np.array([self.parameters.parameter_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])

        # Constant and AR terms
        if self.ar == 0:
            mu = np.transpose(self.ar_matrix)
        elif self.ar == 1:
            mu = np.transpose(self.ar_matrix)*parm[0:-1-self.ma-len(self.X_names)][0]
        else:
            mu = np.matmul(np.transpose(self.ar_matrix),parm[0:-1-self.ma-len(self.X_names)])

        # X terms
        mu = mu + np.matmul(self.X[self.integ+self.max_lag:],parm[self.ma+self.ar:(self.ma+self.ar+len(self.X_names))])

        # MA terms
        if self.ma != 0:
            for t in range(self.max_lag,Y.shape[0]):
                for k in range(0,self.ma):
                        mu[t] += parm[self.ar+k]*(Y[t-1-k]-mu[t-1-k])

        return mu, Y 

    def _mean_prediction(self,mu,Y,h,t_params,X_oos):
        """ Creates a h-step ahead mean prediction

        Parameters
        ----------
        mu : np.array
            The past predicted values

        Y : np.array
            The past data

        h : int
            How many steps ahead for the prediction

        t_params : np.array
            A vector of (transformed) parameters

        X_oos : np.array
            Out of sample X data

        Returns
        ----------
        h-length vector of mean predictions
        """     

        # Create arrays to iteratre over
        Y_exp = Y.copy()
        mu_exp = mu.copy()

        # Loop over h time periods          
        for t in range(0,h):
            new_value = 0
            if self.ar != 0:
                for j in range(0,self.ar):
                    new_value += t_params[j]*Y_exp[-j-1]

            if self.ma != 0:
                for k in range(0,self.ma):
                    if k >= t:
                        new_value += t_params[k+self.ar]*(Y_exp[-k-1]-mu_exp[-k-1])

            # X terms
            new_value += np.matmul(X_oos[t,:],t_params[self.ma+self.ar:(self.ma+self.ar+len(self.X_names))])            
            Y_exp = np.append(Y_exp,[new_value])
            mu_exp = np.append(mu_exp,[0]) # For indexing consistency

        return Y_exp

    def _sim_prediction(self,mu,Y,h,t_params,X_oos,simulations):
        """ Simulates a h-step ahead mean prediction

        Parameters
        ----------
        mu : np.array
            The past predicted values

        Y : np.array
            The past data

        h : int
            How many steps ahead for the prediction

        t_params : np.array
            A vector of (transformed) parameters

        X_oos : np.array
            Out of sample X data

        simulations : int
            How many simulations to perform

        Returns
        ----------
        Matrix of simulations
        """     

        sim_vector = np.zeros([simulations,h])

        for n in range(0,simulations):
            # Create arrays to iteratre over
            Y_exp = Y.copy()
            mu_exp = mu.copy()

            # Loop over h time periods          
            for t in range(0,h):
                new_value = np.random.randn(1)*t_params[-1]

                if self.ar != 0:
                    for j in range(0,self.ar):
                        new_value += t_params[j]*Y_exp[-j-1]

                if self.ma != 0:
                    for k in range(0,self.ma):
                        if k >= t:
                            new_value += t_params[k+self.ar]*(Y_exp[-k-1]-mu_exp[-k-1])

                # X terms
                new_value += np.matmul(X_oos[t,:],t_params[self.ma+self.ar:(self.ma+self.ar+len(self.X_names))])            

                Y_exp = np.append(Y_exp,[new_value])
                mu_exp = np.append(mu_exp,[0]) # For indexing consistency

            sim_vector[n] = Y_exp[-h:]

        return np.transpose(sim_vector)

    def _summarize_simulations(self,mean_values,sim_vector,date_index,h,past_values):
        """ Summarizes a simulation vector and a mean vector of predictions

        Parameters
        ----------
        mean_values : np.array
            Mean predictions for h-step ahead forecasts

        sim_vector : np.array
            N simulation predictions for h-step ahead forecasts

        date_index : pd.DateIndex or np.array
            Dates for the simulations

        h : int
            How many steps ahead are forecast

        past_values : int
            How many past observations to include in the forecast plot

        intervals : Boolean
            Would you like to show prediction intervals for the forecast?
        """         

        error_bars = []
        for pre in range(5,100,5):
            error_bars.append(np.insert([np.percentile(i,pre) for i in sim_vector] - mean_values[-h:],0,0))
        forecasted_values = mean_values[-h-1:]
        plot_values = mean_values[-h-past_values:]
        plot_index = date_index[-h-past_values:]
        return error_bars, forecasted_values, plot_values, plot_index
        
    def neg_loglik(self,beta):
        """ Creates the negative log-likelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        The negative logliklihood of the model
        """     

        mu, Y = self._model(beta)
        return -np.sum(ss.norm.logpdf(Y,loc=mu,scale=self.parameters.parameter_list[-1].prior.transform(beta[-1])))

    def plot_fit(self,**kwargs):
        """ Plots the fit of the model

        Returns
        ----------
        None (plots data and the fit)
        """

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        date_index = self.index[max(self.ar,self.ma):self.data.shape[0]]
        mu, Y = self._model(self.parameters.get_parameter_values())
        plt.plot(date_index,Y,label='Data')
        plt.plot(date_index,mu,label='Filter',c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def plot_predict(self,h=5,past_values=20,intervals=True,oos_data=None,**kwargs):
        """ Plots forecasts with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show prediction intervals for the forecast?

        oos_data : pd.DataFrame
            Data for the variables to be used out of sample (ys can be NaNs)

        Returns
        ----------
        - Plot of the forecast
        """     

        figsize = kwargs.get('figsize',(10,7))

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:
            # Sort/manipulate the out-of-sample data
            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            X_pred = X_oos[:h]

            # Retrieve data, dates and (transformed) parameters
            mu, Y = self._model(self.parameters.get_parameter_values())         
            date_index = self.shift_dates(h)
            t_params = self.transform_parameters()

            # Get mean prediction and simulations (for errors)
            mean_values = self._mean_prediction(mu,Y,h,t_params,X_pred)
            sim_values = self._sim_prediction(mu,Y,h,t_params,X_pred,15000)
            error_bars, forecasted_values, plot_values, plot_index = self._summarize_simulations(mean_values,sim_values,date_index,h,past_values)

            plt.figure(figsize=figsize)
            if intervals == True:
                alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                for count, pre in enumerate(error_bars):
                    plt.fill_between(date_index[-h-1:], forecasted_values-pre, forecasted_values+pre,alpha=alpha[count])            
            plt.plot(plot_index,plot_values)
            plt.title("Forecast for " + self.data_name)
            plt.xlabel("Time")
            plt.ylabel(self.data_name)
            plt.show()

    def predict_is(self,h=5):
        """ Makes dynamic in-sample predictions with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        predictions = []

        for t in range(0,h):
            data1 = self.data_original.iloc[:-(h+t),:]
            data2 = self.data_original.iloc[-h+t:,:]
            x = ARIMAX(ar=self.ar,ma=self.ma,integ=self.integ,formula=self.formula,data=self.data_original[:(-h+t)])
            x.fit(printer=False)
            if t == 0:
                predictions = x.predict(1,oos_data=data2)
            else:
                predictions = pd.concat([predictions,x.predict(h=1,oos_data=data2)])
        
        predictions.rename(columns={0:self.y_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions

    def plot_predict_is(self,h=5,**kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        Returns
        ----------
        - Plot of the forecast against data 
        """     

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        predictions = self.predict_is(h)
        data = self.data[-h:]

        plt.plot(predictions.index,data,label='Data')
        plt.plot(predictions.index,predictions,label='Predictions',c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def predict(self,h=5,oos_data=None):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        oos_data : pd.DataFrame
            Data for the variables to be used out of sample (ys can be NaNs)

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:
            # Sort/manipulate the out-of-sample data
            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            X_pred = X_oos[:h]
            mu, Y = self._model(self.parameters.get_parameter_values())         
            date_index = self.shift_dates(h)
            t_params = self.transform_parameters()

            mean_values = self._mean_prediction(mu,Y,h,t_params,X_pred)
            forecasted_values = mean_values[-h:]
            result = pd.DataFrame(forecasted_values)
            result.rename(columns={0:self.data_name}, inplace=True)
            result.index = date_index[-h:]

            return result






from numpy import abs, exp, power, array, sqrt, pi
from scipy.special import gamma

class Score(object):

    @staticmethod
    def score(y,loc,scale,shape):
        pass

    @staticmethod
    def adj_score(y,loc,scale,shape):
        pass


class ExponentialScore(Score):
    """ Exponential Score

    For dynamic parameter \log{\lambda_{t}}
    """

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def log_lam_score(y,lam):
        return 1 - (lam*y)

    @staticmethod
    def log_lam_adj_score(y,lam):
        return 1 - (lam*y)

    def score(self,y,lam):
        return array([self.log_lam_score(y,lam)])

    def adj_score(self,y,lam):
        return array([self.log_lam_adj_score(y,lam)])


class LaplaceScore(Score):

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def mu_score(y,loc,scale):
        return (y-loc)/float(scale*abs(y-loc))

    @staticmethod
    def mu_information_term(y,loc,scale):
        return -(power(y-loc,2) - power(np.abs(loc-y),2))/(scale*power(np.abs(loc-y),3))

    @staticmethod
    def mu_adj_score(y,loc,scale):
        return LaplaceScore.mu_score(y,loc,scale)/LaplaceScore.mu_information_term(y,loc,scale)

    def score(self,y,loc,scale):
        return array([self.mu_score(y,loc,scale)])

    def adj_score(self,y,loc,scale):
        return array([self.mu_adj_score(y,loc,scale)])


class Normal1Score(Score):
    """ Normal Score (1)

    For dynamic parameters \mu_{t} and \sigma^{2}_{t}
    """

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def mu_score(y,loc,var):
        return 0.5*(y-loc)/var

    @staticmethod
    def mu_adj_score(y,loc,var):
        return (y-loc)

    @staticmethod
    def var_score(y,loc,var):
        return -0.5/var + (0.5*power(y-loc,2))/power(var,2)

    @staticmethod
    def var_adj_score(y,loc,var):
        return power(y-loc,2) - var

    def score(self,y,loc,var):
        return array([self.mu_score(y,loc,var), self.var_score(y,loc,var)])

    def adj_score(self,y,loc,var):
        return array([self.mu_adj_score(y,loc,var), self.var_adj_score(y,loc,var)])


class Normal2Score(Score):
    """ Normal Score (2)

    For dynamic parameters \mu_{t} and \log{\sigma^{2}_{t}}
    """

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def mu_score(y,loc,var):
        return 0.5*(y-loc)/var

    @staticmethod
    def mu_adj_score(y,loc,var):
        return (y-loc)

    @staticmethod
    def log_var_score(y,loc,var):
        return -0.5 + (0.5*power(y-loc,2))/var

    @staticmethod
    def log_var_adj_score(y,loc,var):
        return (power(y-loc,2)/var) - 1

    def score(self,y,loc,var):
        return array([self.mu_score(y,loc,var), self.log_var_score(y,loc,var)])

    def adj_score(self,y,loc,var):
        return array([self.mu_adj_score(y,loc,var), self.log_var_adj_score(y,loc,var)])


class PoissonScore(Score):

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def log_lambda_score(y,lam):
        return (y-lam)

    @staticmethod
    def log_lambda_adj_score(y,lam):
        return (y-lam)/float(lam)

    def score(self,y,lam):
        return array([self.log_lam_score(y,lam)])

    def adj_score(self,y,lam):
        return array([self.log_lam_adj_score(y,lam)])


class SkewtScore(Score):

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def tv_variate_exp(df):
        return (sqrt(df)*gamma((df-1.0)/2.0))/(sqrt(pi)*gamma(df/2.0))

    @staticmethod
    def mu_score(y,loc,scale,shape,skewness):
        if (y-loc)>=0:
            return ((shape+1)/shape)*(y-loc)/(power(skewness*scale,2) + (power(y-loc,2)/shape))
        else:
            return ((shape+1)/shape)*(y-loc)/(power(scale,2) + (power(skewness*(y-loc),2)/shape))

    @staticmethod
    def mu_information_term(y,loc,scale,shape,skewness):
        #return (shape+1)*((power(scale,2)*shape) - power(y-loc,2))/power((power(scale,2)*shape) + power(y-loc,2),2)
        return 1.0

    @staticmethod
    def mu_adj_score(y,loc,scale,shape,skewness):
        return tScore.mu_score(y,loc,scale,shape,skewness) / tScore.mu_information_term(y,loc,scale,shape,skewness)

    def score(self,y,loc,scale,shape,skewness):
        return array([self.mu_score(y,loc,scale,shape,skewness)])

    def adj_score(self,y,loc,scale,shape,skewness):
        return array([self.mu_adj_score(y,loc,scale,shape,skewness)])


class tScore(Score):

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def mu_score(y,loc,scale,shape):
        return ((shape+1)/shape)*(y-loc)/(power(scale,2) + (power(y-loc,2)/shape))

    @staticmethod
    def mu_information_term(y,loc,scale,shape):
        return (shape+1)*((power(scale,2)*shape) - power(y-loc,2))/power((power(scale,2)*shape) + power(y-loc,2),2)

    @staticmethod
    def mu_adj_score(y,loc,scale,shape):
        return tScore.mu_score(y,loc,scale,shape) / tScore.mu_information_term(y,loc,scale,shape)

    def score(self,y,loc,scale,shape):
        return array([self.mu_score(y,loc,scale,shape)])

    def adj_score(self,y,loc,scale,shape):
        return array([self.mu_adj_score(y,loc,scale,shape)])

class BetatScore(Score):

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def mu_score(y,loc,scale,shape):
        try:
            return (((shape+1.0)*power(y-loc,2))/float(shape*exp(scale) + power(y-loc,2))) - 1.0
        except:
            return -1.0

    @staticmethod
    def mu_adj_score(y,loc,scale,shape):
        try:
            return (((shape+1.0)*power(y-loc,2))/float(shape*exp(scale) + power(y-loc,2))) - 1.0
        except:
            return -1.0

    def score(self,y,loc,scale,shape):
        return array([self.mu_score(y,loc,scale,shape)])

    def adj_score(self,y,loc,scale,shape):
        return array([self.mu_adj_score(y,loc,scale,shape)])

class SkewBetatScore(Score):

    def __init__(self):

        super(Score,self).__init__()

    @staticmethod
    def tv_variate_exp(df):
        return (sqrt(df)*gamma((df-1.0)/2.0))/(sqrt(pi)*gamma(df/2.0))

    @staticmethod
    def mu_score(y,loc,scale,shape,skewness):
        try:
            if (y-loc)>=0:
                return (((shape+1.0)*power(y-loc,2))/float(power(skewness,2)*shape*exp(scale) + power(y-loc,2))) - 1.0
            else:
                return (((shape+1.0)*power(y-loc,2))/float(power(skewness,-2)*shape*exp(scale) + power(y-loc,2))) - 1.0    
        except:
            return -1.0

    @staticmethod
    def mu_adj_score(y,loc,scale,shape,skewness):
        try:
            if (y-loc)>=0:
                return (((shape+1.0)*power(y-loc,2))/float(power(skewness,2)*shape*exp(scale) + power(y-loc,2))) - 1.0
            else:
                return (((shape+1.0)*power(y-loc,2))/float(power(skewness,-2)*shape*exp(scale) + power(y-loc,2))) - 1.0    
        except:
            return -1.0

    def score(self,y,loc,scale,shape,skewness):
        return array([self.mu_score(y,loc,scale,shape,skewness)])

    def adj_score(self,y,loc,scale,shape,skewness):
        return array([self.mu_adj_score(y,loc,scale,shape,skewness)])


import scipy.stats as ss
import matplotlib.mlab as mlab
import matplotlib.pyplot as plt
import numpy as np

class q_Normal(object):

    def __init__(self,loc,scale):
        self.loc = loc
        self.scale = scale
        self.param_no = 2

    def sim(self,size): 
        return ss.norm.rvs(loc=self.loc,scale=np.exp(self.scale),size=size)

    def loc_score(self,x):
        return (x-self.loc)/(np.exp(self.scale)**2)

    def scale_score(self,x):
        # For scale = exp(x)
        return np.exp(-2*self.scale)*(x-self.loc)**2 - 1 

    # Indexed scale - for BBVI code
    def score(self,x,index):
        if index == 0:
            return self.loc_score(x)
        elif index == 1:
            return self.scale_score(x)

    def return_param(self,index):
        if index == 0:
            return self.loc
        elif index == 1:
            return self.scale

    def change_param(self,index,value):
        if index == 0:
            self.loc = value
        elif index == 1:
            self.scale = value

    def logpdf(self,x):
        return ss.norm.logpdf(x,loc=self.loc,scale=np.exp(self.scale))

    def plot_pdf(self):
        x = np.linspace(self.loc-np.exp(self.scale)*3.5,self.loc+np.exp(self.scale)*3.5,100)
        plt.plot(x,mlab.normpdf(x,self.loc,np.exp(self.scale)))
        plt.show()



import numpy as np
import sys
if sys.version_info < (3,):
    range = xrange

class SquaredExponential(object):

    def __init__(self, X, l, tau):

        self.X = X.transpose()
        self.Xo = X
        self.l = l
        self.tau = tau

    def K(self):
        K = np.zeros((self.X.shape[0],self.X.shape[0]))

        for i in range(0,self.X.shape[0]):
            for j in range(0,self.X.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-0.5*np.sum(np.power(self.X[i] - self.X[j],2))/np.power(self.l,2))

        K = K + np.identity(K.shape[0])*(10**-10)

        return K

    def Kstar(self,Xstar):
        K = np.zeros((self.X.shape[0],Xstar.shape[0]))

        for i in range(0,self.X.shape[0]):
            for j in range(0,Xstar.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-0.5*np.sum(np.power(self.X[i] - Xstar[j],2))/np.power(self.l,2))

        return K

    def Kstarstar(self,Xstar):
        K = np.zeros((Xstar.shape[0],Xstar.shape[0]))

        for i in range(0,Xstar.shape[0]):
            for j in range(0,Xstar.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-0.5*np.sum(np.power(Xstar[i] - Xstar[j],2))/np.power(self.l,2))

        return K

class ARD(object):

    def __init__(self, X, l, tau):

        self.X = X.transpose()
        self.Xo = X
        self.l = l
        self.tau = tau

    def K(self):
        K = np.zeros((self.X.shape[0],self.X.shape[0]))

        for i in range(0,self.X.shape[0]):
            for j in range(0,self.X.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-0.5*np.sum(np.power(self.X[i] - self.X[j],2)/np.power(self.l,2)))

        K = K + np.identity(K.shape[0])*(10**-10)

        return K

    def Kstar(self,Xstar):
        K = np.zeros((self.X.shape[0],Xstar.shape[0]))

        for i in range(0,self.X.shape[0]):
            for j in range(0,Xstar.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-0.5*np.sum(np.power(self.X[i] - self.X[j],2)/np.power(self.l,2)))

        return K

    def Kstarstar(self,Xstar):
        K = np.zeros((Xstar.shape[0],Xstar.shape[0]))

        for i in range(0,Xstar.shape[0]):
            for j in range(0,Xstar.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-0.5*np.sum(np.power(self.X[i] - self.X[j],2)/np.power(self.l,2)))

        return K


class RationalQuadratic(object):

    def __init__(self, X, l, a, tau):

        self.X = X.transpose()
        self.Xo = X
        self.l = l
        self.a = a
        self.tau = tau

    def K(self):
        K = np.zeros((self.X.shape[0],self.X.shape[0]))

        for i in range(0,self.X.shape[0]):
            for j in range(0,self.X.shape[0]):
                K[i,j] = self.tau*(1.0 + 0.5*np.sum(np.power(self.X[i] - self.X[j],2))/((self.l**2)*self.a))**-self.a

        K = K + np.identity(K.shape[0])*(10**-10)

        return K

    def Kstar(self,Xstar):
        K = np.zeros((self.X.shape[0],Xstar.shape[0]))

        for i in range(0,self.X.shape[0]):
            for j in range(0,Xstar.shape[0]):
                K[i,j] = self.tau*(1.0 + 0.5*np.sum(np.power(self.X[i] - Xstar[j],2))/((self.l**2)*self.a))**-self.a

        return K

    def Kstarstar(self,Xstar):
        K = np.zeros((Xstar.shape[0],Xstar.shape[0]))

        for i in range(0,Xstar.shape[0]):
            for j in range(0,Xstar.shape[0]):
                K[i,j] = self.tau*(1.0 + 0.5*np.sum(np.power(Xstar[i] - Xstar[j],2))/((self.l**2)*self.a))**-self.a

        return K


class OrnsteinUhlenbeck(object):

    def __init__(self, X, l, tau):

        self.X = X.transpose()
        self.Xo = X
        self.l = l
        self.tau = tau

    def K(self):
        K = np.zeros((self.X.shape[0],self.X.shape[0]))

        for i in range(0,self.X.shape[0]):
            for j in range(0,self.X.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-np.sum(np.abs(self.X[i] - self.X[j]))/self.l)

        K = K + np.identity(K.shape[0])*(10**-10)

        return K

    def Kstar(self,Xstar):
        K = np.zeros((self.X.shape[0],Xstar.shape[0]))

        for i in range(0,self.X.shape[0]):
            for j in range(0,Xstar.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-np.sum(np.abs(self.X[i] - Xstar[j]))/self.l)

        return K

    def Kstarstar(self,Xstar):
        K = np.zeros((Xstar.shape[0],Xstar.shape[0]))

        for i in range(0,Xstar.shape[0]):
            for j in range(0,Xstar.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-np.sum(np.abs(Xstar[i] - Xstar[j]))/self.l)

        return K

class Periodic(object):

    def __init__(self, X, l, tau):

        self.X = X.transpose()
        self.Xo = X
        self.l = l
        self.tau = tau

    def K(self):
        K = np.zeros((self.X.shape[0],self.X.shape[0]))

        for i in range(0,self.X.shape[0]):
            for j in range(0,self.X.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-np.sum(np.power(-2.0*np.sin(self.X[i] - self.X[j]),2))/self.l)

        K = K + np.identity(K.shape[0])*(10**-10)

        return K

    def Kstar(self,Xstar):
        K = np.zeros((self.X.shape[0],Xstar.shape[0]))

        for i in range(0,self.X.shape[0]):
            for j in range(0,Xstar.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-np.sum(np.power(-2.0*np.sin(self.X[i] - self.X[j]),2))/self.l)

        return K

    def Kstarstar(self,Xstar):
        K = np.zeros((Xstar.shape[0],Xstar.shape[0]))

        for i in range(0,Xstar.shape[0]):
            for j in range(0,Xstar.shape[0]):
                # Due to Toeplitz structure
                if i == j:
                    K[i,j] = self.tau
                else:
                    K[i,j] = self.tau*np.exp(-np.sum(np.power(-2.0*np.sin(self.X[i] - self.X[j]),2))/self.l)

        return K




import copy
import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
import matplotlib.pyplot as plt
import seaborn as sns

from .. import inference as ifr
from .. import distributions as dst
from .. import tsm as tsm
from .. import data_check as dc

from .kalman import *

class LLT(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** LOCAL LINEAR TREND MODEL ****

    Parameters
    ----------
    data : pd.DataFrame or np.array
        Field to specify the time series data that will be used.

    integ : int (default : 0)
        Specifies how many time to difference the time series.

    target : str (pd.DataFrame) or int (np.array)
        Specifies which column name or array index to use. By default, first
        column/array will be selected as the dependent variable.
    """

    def __init__(self,data,integ=0,target=None):

        # Initialize TSM object
        super(LLT,self).__init__('LLT')

        # Parameters
        self.integ = integ
        self.param_no = 3
        self.max_lag = 0
        self._param_hide = 0 # Whether to cutoff variance parameters from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"
        self.model_name = "LLT"
        self.multivariate_model = False

        # Format the data
        self.data, self.data_name, self.is_pandas, self.index = dc.data_check(data,target)
        self.data_original = self.data

        # Difference data
        for order in range(self.integ):
            self.data = np.diff(self.data)
            self.data_name = "Differenced " + self.data_name

        self._create_parameters()

    def _create_parameters(self):
        """ Creates model parameters

        Returns
        ----------
        None (changes model attributes)
        """

        self.parameters.add_parameter('Sigma^2 irregular',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))
        self.parameters.add_parameter('Sigma^2 level',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))
        self.parameters.add_parameter('Sigma^2 trend',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))

    def _forecast_model(self,beta,h):
        """ Creates forecasted states and variances

        Parameters
        ----------
        beta : np.ndarray
            Contains untransformed starting values for parameters

        Returns
        ----------
        a : np.ndarray
            Forecasted states

        P : np.ndarray
            Variance of forecasted states
        """     

        T, Z, R, Q, H = self._ss_matrices(beta)
        return univariate_kalman_fcst(self.data,Z,H,T,Q,R,0.0,h)

    def _model(self,data,beta):
        """ Creates the structure of the model

        Parameters
        ----------
        data : np.array
            Contains the time series

        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        a,P,K,F,v : np.array
            Filted states, filtered variances, Kalman gains, F matrix, residuals
        """     

        T, Z, R, Q, H = self._ss_matrices(beta)

        return univariate_kalman(data,Z,H,T,Q,R,0.0)

    def _ss_matrices(self,beta):
        """ Creates the state space matrices required

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        T, Z, R, Q, H : np.array
            State space matrices used in KFS algorithm
        """     

        T = np.identity(2)
        T[0][1] = 1
        
        Z = np.zeros(2)
        Z[0] = 1

        R = np.identity(2)
        Q = np.identity(2)
        H = np.identity(1)*self.parameters.parameter_list[0].prior.transform(beta[0])
        Q[0][0] = self.parameters.parameter_list[1].prior.transform(beta[1])
        Q[1][1] = self.parameters.parameter_list[2].prior.transform(beta[2])

        return T, Z, R, Q, H

    def neg_loglik(self,beta):
        """ Creates the negative log likelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        The negative log logliklihood of the model
        """         
        _, _, _, F, v = self._model(self.data,beta)
        loglik = 0.0
        for i in range(0,self.data.shape[0]):
            loglik += np.linalg.slogdet(F[:,:,i])[1] + np.dot(v[i],np.dot(np.linalg.pinv(F[:,:,i]),v[i]))
        return -(-((self.data.shape[0]/2)*np.log(2*np.pi))-0.5*loglik.T[0].sum())

    def plot_predict(self,h=5,past_values=20,intervals=True,**kwargs):      
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show 95% prediction intervals for the forecast?

        Returns
        ----------
        - Plot of the forecast
        """     

        figsize = kwargs.get('figsize',(10,7))

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:
            # Retrieve data, dates and (transformed) parameters         
            a, P = self._forecast_model(self.parameters.get_parameter_values(),h)
            date_index = self.shift_dates(h)
            plot_values = a[0][-h-past_values:]
            forecasted_values = a[0][-h:]
            lower = forecasted_values - 1.98*np.power(P[0][0][-h:] + self.parameters.parameter_list[0].prior.transform(self.parameters.get_parameter_values()[0]),0.5)
            upper = forecasted_values + 1.98*np.power(P[0][0][-h:] + self.parameters.parameter_list[0].prior.transform(self.parameters.get_parameter_values()[0]),0.5)
            lower = np.append(plot_values[-h-1],lower)
            upper = np.append(plot_values[-h-1],upper)

            plot_index = date_index[-h-past_values:]

            plt.figure(figsize=figsize)
            if intervals == True:
                plt.fill_between(date_index[-h-1:], lower, upper, alpha=0.2)            

            plt.plot(plot_index,plot_values)
            plt.title("Forecast for " + self.data_name)
            plt.xlabel("Time")
            plt.ylabel(self.data_name)
            plt.show()

    def plot_fit(self,intervals=True,**kwargs):
        """ Plots the fit of the model

        Parameters
        ----------
        intervals : Boolean
            Whether to plot 95% confidence interval of states

        Returns
        ----------
        None (plots data and the fit)
        """

        figsize = kwargs.get('figsize',(10,7))
        series_type = kwargs.get('series_type','Smoothed')

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:
            date_index = copy.deepcopy(self.index)
            date_index = date_index[self.integ:self.data_original.shape[0]+1]

            if series_type == 'Smoothed':
                mu, V= self.smoothed_state(self.data,self.parameters.get_parameter_values())
            elif series_type == 'Filtered':
                mu, V, _, _, _ = self._model(self.data,self.parameters.get_parameter_values())
            else:
                mu, V = self.smoothed_state(self.data,self.parameters.get_parameter_values())

            mu0 = mu[0][:-1]
            mu1 = mu[1][:-1]
            Vlev = V[0][0][:-1]
            Vtrend = V[0][1][:-1]   

            plt.figure(figsize=figsize) 
            
            plt.subplot(2, 2, 1)
            plt.title(self.data_name + " Raw and " + series_type)   

            if intervals == True:
                alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                plt.fill_between(date_index[5:], mu0[5:] + 1.98*np.sqrt(Vlev[5:]), mu0[5:] - 1.98*np.sqrt(Vlev[5:]), alpha=0.15,label='95% C.I.')   

            plt.plot(date_index,self.data,label='Data')
            plt.plot(date_index,mu0,label=series_type,c='black')
            plt.legend(loc=2)
            
            plt.subplot(2, 2, 2)

            if intervals == True:
                alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                plt.fill_between(date_index[5:], mu0[5:] + 1.98*np.sqrt(Vlev[5:]), mu0[5:] - 1.98*np.sqrt(Vlev[5:]), alpha=0.15,label='95% C.I.')   

            plt.title(self.data_name + " Local Level")  
            plt.plot(date_index,mu0,label='Local Level')
            plt.legend(loc=2)
            
            plt.subplot(2, 2, 3)

            if intervals == True:
                alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                plt.fill_between(date_index[5:], mu1[5:] + 1.98*np.sqrt(Vtrend[5:]), mu1[5:] - 1.98*np.sqrt(Vtrend[5:]), alpha=0.15,label='95% C.I.')   

            plt.title(self.data_name + " Trend")    
            plt.plot(date_index,mu1,label='Stochastic Trend')
            plt.legend(loc=2)

            plt.subplot(2, 2, 4)
            plt.title("Measurement Noise")  
            plt.plot(date_index[1:self.data.shape[0]],self.data[1:self.data.shape[0]]-mu0[1:self.data.shape[0]],label='Irregular term')
            plt.show()              

    def predict(self,h=5):      
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predictions
        """     

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:
            # Retrieve data, dates and (transformed) parameters         
            a, P = self._forecast_model(self.parameters.get_parameter_values(),h)
            date_index = self.shift_dates(h)
            forecasted_values = a[0][-h:]

            result = pd.DataFrame(forecasted_values)
            result.rename(columns={0:self.data_name}, inplace=True)
            result.index = date_index[-h:]

            return result

    def predict_is(self,h=5):
        """ Makes dynamic in-sample predictions with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        predictions = []

        for t in range(0,h):
            x = LLT(integ=self.integ,data=self.data_original[:(-h+t)])
            x.fit(printer=False)
            if t == 0:
                predictions = x.predict(1)
            else:
                predictions = pd.concat([predictions,x.predict(1)])
        
        predictions.rename(columns={0:self.data_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions

    def plot_predict_is(self,h=5,**kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        Returns
        ----------
        - Plot of the forecast against data 
        """     

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        predictions = self.predict_is(h)
        data = self.data[-h:]
        plt.plot(predictions.index,data,label='Data')
        plt.plot(predictions.index,predictions,label='Predictions',c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def simulation_smoother(self,beta):
        """ Koopman's simulation smoother - simulates from states given
        model parameters and observations

        Parameters
        ----------

        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        - A simulated state evolution
        """         

        T, Z, R, Q, H = self._ss_matrices(beta)

        # Generate e_t+ and n_t+
        rnd_h = np.random.normal(0,np.sqrt(H),self.data.shape[0]+1)
        q_dist = ss.multivariate_normal([0.0, 0.0], Q)
        rnd_q = q_dist.rvs(self.data.shape[0]+1)

        # Generate a_t+ and y_t+
        a_plus = np.zeros((T.shape[0],self.data.shape[0]+1)) 
        a_plus[0,0] = np.mean(self.data[0:5])
        y_plus = np.zeros(self.data.shape[0])

        for t in range(0,self.data.shape[0]+1):
            if t == 0:
                a_plus[:,t] = np.dot(T,a_plus[:,t]) + rnd_q[t,:]
                y_plus[t] = np.dot(Z,a_plus[:,t]) + rnd_h[t]
            else:
                if t != self.data.shape[0]:
                    a_plus[:,t] = np.dot(T,a_plus[:,t-1]) + rnd_q[t,:]
                    y_plus[t] = np.dot(Z,a_plus[:,t]) + rnd_h[t]

        alpha_hat,_ = self.smoothed_state(self.data,beta)
        alpha_hat_plus,_ = self.smoothed_state(y_plus,beta)
        alpha_tilde = alpha_hat - alpha_hat_plus + a_plus
    
        return alpha_tilde

    def smoothed_state(self,data,beta):
        """ Creates the negative log marginal likelihood of the model

        Parameters
        ----------

        data : np.array
            Data to be smoothed

        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        - Smoothed states
        """         

        T, Z, R, Q, H = self._ss_matrices(beta)
        alpha, V = univariate_KFS(data,Z,H,T,Q,R,0.0)
        return alpha, V


import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
import scipy.special as sp
import matplotlib.pyplot as plt
import seaborn as sns
from patsy import dmatrices, dmatrix, demo_data

from ..parameter import Parameter, Parameters
from .. import inference as ifr
from .. import tsm as tsm
from .. import distributions as dst
from .. import data_check as dc

from .scores import *

class GASReg(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** GENERALIZED AUTOREGRESSIVE SCORE (GAS) REGRESSION MODELS ****

    Parameters
    ----------

    formula : string
        patsy string describing the regression

    data : pd.DataFrame or np.array
        Field to specify the data that will be used

    family : GAS family object
        Which distribution to use, e.g. GASNormal()

    """

    def __init__(self, formula, data, family):

        # Initialize TSM object     
        super(GASReg,self).__init__('GASReg')

        # Parameters
        self.max_lag = 0
        self._param_hide = 0 # Whether to cutoff variance parameters from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"
        self.multivariate_model = False
        self.skewness = False

        # Format the data
        self.is_pandas = True # This is compulsory for this model type
        self.data_original = data
        self.formula = formula
        self.y, self.X = dmatrices(formula, data)
        self.param_no = self.X.shape[1]
        self.y_name = self.y.design_info.describe()
        self.data_name = self.y_name
        self.X_names = self.X.design_info.describe().split(" + ")
        self.y = np.array([self.y]).ravel()
        self.data = self.y
        self.X = np.array([self.X])[0]
        self.index = data.index
        self.initial_values = np.zeros(self.param_no)

        self._create_model_matrices()
        self._create_parameters()

        self.family = family
        
        self.model_name2, self.link, self.scale, self.shape, self.skewness, self.mean_transform = self.family.setup()
        self.model_name = self.model_name2 + " Regression"

        for no, i in enumerate(self.family.build_parameters()):
            self.parameters.add_parameter(i[0],i[1],i[2])
            self.parameters.parameter_list[no+self.param_no].start = i[3]

        self.param_no += len(self.family.build_parameters())

    def _create_model_matrices(self):
        """ Creates model matrices/vectors

        Returns
        ----------
        None (changes model attributes)
        """

        self.model_Y = self.data
        self.model_scores = np.zeros((self.X.shape[1],self.model_Y.shape[0]+1))

    def _create_parameters(self):
        """ Creates model parameters

        Returns
        ----------
        None (changes model attributes)
        """

        for parm in range(self.param_no):
            self.parameters.add_parameter('Scale ' + self.X_names[parm],ifr.Uniform(transform='exp'),dst.q_Normal(0,3))
            self.parameters.parameter_list[parm].start = -7.0

    def _get_scale_and_shape(self,parm):
        """ Obtains appropriate model scale and shape parameters

        Parameters
        ----------
        parm : np.array
            Transformed parameter vector

        Returns
        ----------
        None (changes model attributes)
        """

        if self.scale is True:
            if self.shape is True:
                model_shape = parm[-1]  
                model_scale = parm[-2]
            else:
                model_shape = 0
                model_scale = parm[-1]
        else:
            model_scale = 0
            model_shape = 0 
        if self.skewness is True:
            model_skewness = parm[-3]
        else:
            model_skewness = 0
        return model_scale, model_shape, model_skewness

    def _model(self,beta):
        """ Creates the structure of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        theta : np.array
            Contains the predicted values for the time series

        Y : np.array
            Contains the length-adjusted time series (accounting for lags)

        scores : np.array
            Contains the scores for the time series
        """

        parm = np.array([self.parameters.parameter_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        coefficients = np.zeros((self.X.shape[1],self.model_Y.shape[0]+1))
        coefficients[:,0] = self.initial_values
        theta = np.zeros(self.model_Y.shape[0]+1)
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)

        # Loop over time series
        for t in range(0,self.model_Y.shape[0]):
            theta[t] = np.dot(self.X[t],coefficients[:,t])
            self.model_scores[:,t] = self.family.reg_score_function(self.X[t],self.model_Y[t],self.link(theta[t]),model_scale,model_shape,model_skewness)
            coefficients[:,t+1] = coefficients[:,t] + parm[0:self.X.shape[1]]*self.model_scores[:,t] 
        return theta[:-1], self.model_Y, self.model_scores[:-1], coefficients

    def neg_loglik(self,beta):
        theta, Y, scores,_ = self._model(beta)
        parm = np.array([self.parameters.parameter_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)
        return self.family.neg_loglikelihood(Y,self.link(theta),model_scale,model_shape,model_skewness)

    def plot_fit(self,**kwargs):
        """ Plots the fit of the model

        Notes
        ----------
        Intervals are bootstrapped as follows: take the filtered values from the
        algorithm (thetas). Use these thetas to generate a pseudo data stream from
        the measurement density. Use the GAS algorithm and estimated parameters to
        filter the pseudo data. Repeat this N times. 

        Returns
        ----------
        None (plots data and the fit)
        """

        figsize = kwargs.get('figsize',(10,7))

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:

            date_index = self.index.copy()
            mu, Y, scores, coefficients = self._model(self.parameters.get_parameter_values())

            if self.model_name2 == "Exponential GAS":
                values_to_plot = 1.0/self.link(mu)
            elif self.model_name2 == "Skewt GAS":
                t_params = self.transform_parameters()
                model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_params)
                m1 = (np.sqrt(model_shape)*sp.gamma((model_shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(model_shape/2.0))
                additional_loc = (model_skewness - (1.0/model_skewness))*model_scale*m1
                values_to_plot = mu + additional_loc
            else:
                values_to_plot = self.link(mu)

            plt.figure(figsize=figsize) 
            
            plt.subplot(len(self.X_names)+1, 1, 1)
            plt.title(self.y_name + " Filtered")
            plt.plot(date_index,Y,label='Data')
            plt.plot(date_index,values_to_plot,label='GAS Filter',c='black')
            plt.legend(loc=2)

            for coef in range(0,len(self.X_names)):
                plt.subplot(len(self.X_names)+1, 1, 2+coef)
                plt.title("Beta " + self.X_names[coef]) 
                plt.plot(date_index,coefficients[coef,0:-1],label='Coefficient')
                plt.legend(loc=2)               

            plt.show()          
    
    def plot_predict(self,h=5,past_values=20,intervals=True,oos_data=None,**kwargs):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show prediction intervals for the forecast?

        oos_data : pd.DataFrame
            Data for the variables to be used out of sample (ys can be NaNs)

        Returns
        ----------
        - Plot of the forecast
        """     

        figsize = kwargs.get('figsize',(10,7))

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:

            # Sort/manipulate the out-of-sample data
            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            X_pred = X_oos[:h]
            date_index = self.shift_dates(h)
            _, _, _, coefficients = self._model(self.parameters.get_parameter_values()) 
            coefficients_star = coefficients.T[-1]
            theta_pred = np.dot(np.array([coefficients_star]), X_pred.T)[0]
            t_params = self.transform_parameters()
            model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_params)

            # Measurement prediction intervals
            rnd_value = self.family.draw_variable(self.link(theta_pred),model_scale,model_shape,model_skewness,[1500,theta_pred.shape[0]])

            if self.model_name2 == "Skewt GAS":
                model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_params)
                m1 = (np.sqrt(model_shape)*sp.gamma((model_shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(model_shape/2.0))
                theta_pred += (model_skewness - (1.0/model_skewness))*model_scale*m1 

            error_bars = []
            for pre in range(5,100,5):
                error_bars.append(np.insert([np.percentile(i,pre) for i in rnd_value.T] - self.link(theta_pred),0,0))

            plot_values = np.append(self.y,self.link(theta_pred))
            plot_values = plot_values[-h-past_values:]
            forecasted_values = np.append(self.y[-1],self.link(theta_pred))
            plot_index = date_index[-h-past_values:]

            plt.figure(figsize=figsize)
            if intervals == True:
                alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                for count, pre in enumerate(error_bars):
                    plt.fill_between(date_index[-h-1:], forecasted_values-pre, forecasted_values+pre,alpha=alpha[count])            
            
            plt.plot(plot_index,plot_values)
            plt.title("Forecast for " + self.y_name)
            plt.xlabel("Time")
            plt.ylabel(self.y_name)
            plt.show()

    def plot_predict_is(self,h=5,**kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        Returns
        ----------
        - Plot of the forecast against data 
        """     

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        predictions = self.predict_is(h)
        data = self.data[-h:]

        plt.plot(predictions.index,data,label='Data')
        plt.plot(predictions.index,predictions,label='Predictions',c='black')
        plt.title(self.y_name)
        plt.legend(loc=2)   
        plt.show()          

    def predict(self,h=5,oos_data=None):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        oos_data : pd.DataFrame
            Data for the variables to be used out of sample (ys can be NaNs)

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:

            # Sort/manipulate the out-of-sample data
            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            X_pred = X_oos[:h]
            date_index = self.shift_dates(h)
            _, _, _, coefficients = self._model(self.parameters.get_parameter_values()) 
            coefficients_star = coefficients.T[-1]
            theta_pred = np.dot(np.array([coefficients_star]), X_pred.T)[0]

            if self.model_name2 == "Skewt GAS":
                t_params = self.transform_parameters()
                model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_params)
                m1 = (np.sqrt(model_shape)*sp.gamma((model_shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(model_shape/2.0))
                theta_pred = theta_pred + (model_skewness - (1.0/model_skewness))*model_scale*m1 

            result = pd.DataFrame(self.link(theta_pred))
            result.rename(columns={0:self.y_name}, inplace=True)
            result.index = date_index[-h:]

            return result

    def predict_is(self,h=5):
        """ Makes dynamic in-sample predictions with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        predictions = []

        for t in range(0,h):
            data1 = self.data_original.iloc[:-(h+t),:]
            data2 = self.data_original.iloc[-h+t:,:]
            x = GASReg(formula=self.formula,data=self.data_original[:(-h+t)],family=self.family)
            x.fit(printer=False)
            if t == 0:
                predictions = x.predict(1,oos_data=data2)
            else:
                predictions = pd.concat([predictions,x.predict(h=1,oos_data=data2)])
        
        predictions.rename(columns={0:self.y_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions


import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
from scipy.stats import multivariate_normal
import matplotlib.pyplot as plt
import seaborn as sns
from patsy import dmatrices, dmatrix, demo_data

from .. import arma
from .. import inference as ifr
from .. import distributions as dst
from .. import output as op
from .. import tests as tst
from .. import tsm as tsm
from .. import data_check as dc

from .kernels import *
from .gp import *

class GPR(GP):
    """ Inherits time series methods from TSM class.

    **** GAUSSIAN PROCESS REGRESSION ****

    Parameters
    ----------
    formula : string
        Patsy regression

    data : pd.DataFrame or np.array
        Field to specify the time series data that will be used.

    kernel_type : str
        One of SE (SquaredExponential), OU (Ornstein-Uhlenbeck), RQ
        (RationalQuadratic), Periodic, ARD. Defines kernel choice for GP-NARX.
    """

    def __init__(self,formula,data,kernel_type='SE'):

        # Initialize TSM object
        super(GP,self).__init__('GP')

        # Parameters
        self._param_hide = 0 # Whether to cutoff variance parameters from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"
        self.multivariate_model = False

        # Format the data
        self.max_lag = 0
        self.is_pandas = True # This is compulsory for this model type
        self.data_original = data
        self.formula = formula
        self.y, self.X = dmatrices(formula, data)
        self.param_no = self.X.shape[1]
        self.y_name = self.y.design_info.describe()
        self.data_name = self.y_name
        self.X_names = self.X.design_info.describe().split(" + ")
        self.y = np.array([self.y]).ravel()
        self.data = self.y
        self.X = np.array([self.X])[0].T
        self.index = data.index
        self.X_original = self.X.copy()

        if kernel_type == 'ARD':
            self.param_no += 2 
        elif kernel_type == 'RQ':
            self.param_no = 3 + 1
        else:
            self.param_no = 3

        self.model_name = 'GPR'
        self.kernel_type = kernel_type

        # Apply normalization
        self.data = self.y.copy()
        self._norm_mean = np.mean(self.data)
        self._norm_std = np.std(self.data)  
        self.data = (self.data - self._norm_mean) / self._norm_std

        for regressor in range(self.X.shape[0]):
            self.X[regressor] = (self.X[regressor] - np.mean(self.X[regressor])) / np.std(self.X[regressor])

        self._create_parameters()

    def _create_parameters(self):
        """ Creates model parameters

        Returns
        ----------
        None (changes model attributes)
        """

        self.parameters.add_parameter('Noise Sigma^2',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))

        if self.kernel_type == 'ARD':
            self.kernel = ARD(self.X,np.ones(len(self.X_names)),1)

            for reg in range(self.X.shape[0]):
                self.parameters.add_parameter('l-' + self.X_names[reg],ifr.Uniform(transform='exp'),dst.q_Normal(0,3))

        else:
            self.parameters.add_parameter('l',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))

            if self.kernel_type == 'SE':
                self.kernel = SquaredExponential(self.X,1,1)
            elif self.kernel_type == 'OU':
                self.kernel = OrnsteinUhlenbeck(self.X,1,1)
            elif self.kernel_type == 'Periodic':
                self.kernel = Periodic(self.X,1,1)
            elif self.kernel_type == 'RQ':
                self.parameters.add_parameter('alpha',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))
                self.kernel = RationalQuadratic(self.X,1,1,1)

        self.parameters.add_parameter('tau',ifr.Uniform(transform='exp'),dst.q_Normal(0,3))

    def _start_params(self,beta):
        """ Transforms parameters for use in kernels

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        None (changes data in self.kernel)
        """

        if self.kernel_type == 'ARD':
            for reg in range(self.X.shape[0]):
                self.kernel.l[reg] = self.parameters.parameter_list[1+reg].prior.transform(beta[1+reg])
            self.kernel.tau = self.parameters.parameter_list[-1].prior.transform(beta[-1])  
        elif self.kernel_type == 'RQ':
            self.kernel.l = self.parameters.parameter_list[1].prior.transform(beta[1])
            self.kernel.a = self.parameters.parameter_list[2].prior.transform(beta[2])          
            self.kernel.tau = self.parameters.parameter_list[3].prior.transform(beta[3])    
        else:
            self.kernel.l = self.parameters.parameter_list[1].prior.transform(beta[1])
            self.kernel.tau = self.parameters.parameter_list[2].prior.transform(beta[2])            

    def plot_oned(self,figsize=(15,10),intervals=True):
        if self.parameters.estimated is True and len(self.X_names) == 1:
            X_sort = self.X.copy()
            for parameter in range(X_sort.shape[0]):
                X_sort[parameter] = np.sort(self.X[parameter])

            Kstar = self.kernel.Kstar(np.transpose(np.array(X_sort)))
            L = self._L(self.parameters.get_parameter_values())
            output =  np.dot(np.transpose(Kstar),self._alpha(L))

            posterior = multivariate_normal(output,self.variance_values(self.parameters.get_parameter_values()),allow_singular=True)
            simulations = 500
            sim_vector = np.zeros([simulations,len(output)])

            for i in range(simulations):
                sim_vector[i] = posterior.rvs()

            error_bars = []
            for pre in range(5,100,5):
                error_bars.append([(np.percentile(i,pre)*self._norm_std + self._norm_mean) for i in sim_vector.transpose()] 
                    - (output*self._norm_std + self._norm_mean))

            plt.figure(figsize=figsize)
            plt.plot(X_sort[0]*np.std(self.X_original[0]) + np.mean(self.X[0]),output*self._norm_std + self._norm_mean)

            if intervals == True:
                alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                for count, pre in enumerate(error_bars):
                    plt.fill_between(X_sort[0]*np.std(self.X_original[0]) + np.mean(self.X[0]), (output*self._norm_std + self._norm_mean)-pre, 
                        (output*self._norm_std + self._norm_mean)+pre,alpha=alpha[count])       

            plt.title(self.y_name + "=" + "f(" + self.X_names[0] + ")")
            plt.xlabel(self.X_names[0])
            plt.ylabel(self.y_name)
            plt.show()
        elif self.parameters.estimated is True and len(self.X_names) != 1:
            raise ValueError("Not a 1-D problem")
        else:
            raise ValueError("No parameters have been estimated yet")

    def plot_fit(self,intervals=True,**kwargs):
        """ Plots the fit of the Gaussian process model to the data

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        intervals : Boolean
            Whether to plot uncertainty intervals or not

        Returns
        ----------
        None (plots the fit of the function)
        """

        figsize = kwargs.get('figsize',(10,7))

        date_index = self.index
        expectation = self.expected_values(self.parameters.get_parameter_values())
        posterior = multivariate_normal(expectation,self.variance_values(self.parameters.get_parameter_values()),allow_singular=True)
        simulations = 500
        sim_vector = np.zeros([simulations,len(expectation)])

        for i in range(simulations):
            sim_vector[i] = posterior.rvs()

        error_bars = []
        for pre in range(5,100,5):
            error_bars.append([(np.percentile(i,pre)*self._norm_std + self._norm_mean) for i in sim_vector.transpose()] 
                - (expectation*self._norm_std + self._norm_mean))

        plt.figure(figsize=figsize) 

        plt.subplot(2, 2, 1)
        plt.title(self.data_name + " Raw")  
        plt.plot(date_index,self.data*self._norm_std + self._norm_mean,'k')

        plt.subplot(2, 2, 2)
        plt.title(self.data_name + " Raw and Expected") 
        plt.plot(date_index,self.data*self._norm_std + self._norm_mean,'k',alpha=0.2)
        plt.plot(date_index,self.expected_values(self.parameters.get_parameter_values())*self._norm_std + self._norm_mean,'b')

        plt.subplot(2, 2, 3)
        plt.title(self.data_name + " Raw and Expected (with intervals)")    

        if intervals == True:
            alpha =[0.15*i/float(100) for i in range(50,12,-2)]
            for count, pre in enumerate(error_bars):
                plt.fill_between(date_index, (expectation*self._norm_std + self._norm_mean)-pre, 
                    (expectation*self._norm_std + self._norm_mean)+pre,alpha=alpha[count])      

        plt.plot(date_index,self.data*self._norm_std + self._norm_mean,'k',alpha=0.2)
        plt.plot(date_index,self.expected_values(self.parameters.get_parameter_values())*self._norm_std + self._norm_mean,'b')

        plt.subplot(2, 2, 4)

        plt.title("Expected " + self.data_name + " (with intervals)")   

        if intervals == True:
            alpha =[0.15*i/float(100) for i in range(50,12,-2)]
            for count, pre in enumerate(error_bars):
                plt.fill_between(date_index, (expectation*self._norm_std + self._norm_mean)-pre, 
                    (expectation*self._norm_std + self._norm_mean)+pre,alpha=alpha[count])      

        plt.plot(date_index,self.expected_values(self.parameters.get_parameter_values())*self._norm_std + self._norm_mean,'b')

        plt.show()


import numpy as np
from math import exp
import sys
if sys.version_info < (3,):
    range = xrange

from scipy.stats import multivariate_normal

class MetropolisHastings(object):
    """ RANDOM-WALK METROPOLIS-HASTINGS MCMC

    Parameters
    ----------
    posterior : function
        A posterior function

    scale : float
        The scale for the random walk

    nsims : int
        The number of simulations to perform

    initials : np.array
        Where to start the MCMC chain

    cov_matrix : np.array
        (optional) A covariance matrix for the random walk
    
    model_object : TSM object
        A model object (for use in SPDK sampling)
    """

    def __init__(self,posterior,scale,nsims,initials,cov_matrix=None,model_object=None):
        self.posterior = posterior
        self.scale = scale
        self.nsims = nsims
        self.initials = initials
        self.param_no = self.initials.shape[0]
        self.phi = np.zeros([nsims,self.param_no])
        self.phi[0] = self.initials
        
        if cov_matrix is None:
            self.cov_matrix = np.identity(self.param_no)
        else:
            self.cov_matrix = cov_matrix

        if model_object is not None:
            self.model = model_object

    @staticmethod
    def tune_scale(acceptance,scale):
        """ Tunes scale for M-H algorithm

        Parameters
        ----------
        acceptance : float
            The most recent acceptance rate

        scale : float
            The current scale parameter

        Returns
        ----------
        scale : float
            An adjusted scale parameter
        """     

        if acceptance > 0.8:
            scale *= 2.0
        elif acceptance <= 0.8 and acceptance > 0.4:
            scale *= 1.3            
        elif acceptance < 0.234 and acceptance > 0.1:
            scale *= (1/1.3)
        elif acceptance <= 0.1 and acceptance > 0.05:
            scale *= 0.4
        elif acceptance <= 0.05 and acceptance > 0.01:
            scale *= 0.2
        elif acceptance <= 0.01:
            scale *= 0.1
        return scale        

    def sample(self):
        """ Sample from M-H algorithm

        Returns
        ----------
        chain : np.array
            Chains for each parameter

        mean_est : np.array
            Mean values for each parameter

        median_est : np.array
            Median values for each parameter

        upper_95_est : np.array
            Upper 95% credibility interval for each parameter

        lower_95_est : np.array
            Lower 95% credibility interval for each parameter           
        """     

        acceptance = 1
        finish = 0

        while (acceptance < 0.234 or acceptance > 0.4) or finish == 0:

            # If acceptance is in range, proceed to sample, else continue tuning
            if not (acceptance < 0.234 or acceptance > 0.4):
                finish = 1
                print("")
                print("Tuning complete! Now sampling.")
                sims_to_do = self.nsims
            else:
                sims_to_do = int(self.nsims/2) # For acceptance rate tuning

            # Holds data on acceptance rates and uniform random numbers
            a_rate = np.zeros([sims_to_do,1])
            crit = np.random.rand(sims_to_do,1)
            post = multivariate_normal(np.zeros(self.param_no),self.cov_matrix)
            rnums = post.rvs()*self.scale
            for k in range(1,sims_to_do): 
                rnums = np.vstack((rnums,post.rvs()*self.scale))

            old_lik = -self.posterior(self.phi[0]) # Initial posterior

            # Sampling time!
            for i in range(1,sims_to_do):
                phi_prop = self.phi[i-1] + rnums[i]
                post_prop = -self.posterior(phi_prop)
                lik_rat = exp(post_prop - old_lik)

                if crit[i] < lik_rat:
                    self.phi[i] = phi_prop
                    a_rate[i] = 1
                    old_lik = post_prop
                else:
                    self.phi[i] = self.phi[i-1]

            acceptance = a_rate.sum()/a_rate.shape[0]
            self.scale = self.tune_scale(acceptance,self.scale)

            print("Acceptance rate of Metropolis-Hastings is " + str(acceptance))

        chain = np.array([self.phi[i][0] for i in range(0,self.phi.shape[0])])
        for m in range(1,self.param_no):
            chain = np.vstack((chain,[self.phi[i][m] for i in range(0,self.phi.shape[0])]))

        mean_est = np.array([np.mean(np.array([self.phi[i][j] for i in range(0,self.phi.shape[0])])) for j in range(self.param_no)])
        median_est = np.array([np.median(np.array([self.phi[i][j] for i in range(0,self.phi.shape[0])])) for j in range(self.param_no)])
        upper_95_est = np.array([np.percentile(np.array([self.phi[i][j] for i in range(0,self.phi.shape[0])]),95) for j in range(self.param_no)])
        lower_95_est = np.array([np.percentile(np.array([self.phi[i][j] for i in range(0,self.phi.shape[0])]),5) for j in range(self.param_no)])

        return chain, mean_est, median_est, upper_95_est, lower_95_est


from math import exp, log, tanh
import numpy as np

def ilogit(x):
    return 1/(1+np.exp(-x))

def logit(x):
    return np.log(x) - np.log(1 - x)

def transform_define(transform):
    if transform == 'tanh':
        return np.tanh
    elif transform == 'exp':
        return np.exp
    elif transform == 'logit':
        return ilogit
    elif transform is None:
        return np.array
    else:
        return None

def itransform_define(transform):
    if transform == 'tanh':
        return np.arctanh
    elif transform == 'exp':
        return np.log
    elif transform == 'logit':
        return logit
    elif transform is None:
        return np.array
    else:
        return None

class Normal(object):

    def __init__(self,mu0,sigma0,transform=None):
        self.mu0 = mu0
        self.sigma0 = sigma0
        self.transform_name = transform     
        self.transform = transform_define(transform)
        self.itransform = itransform_define(transform)

    def logpdf(self,mu):
        if self.transform is not None:
            mu = self.transform(mu)     
        return -log(float(self.sigma0)) - (0.5*(mu-self.mu0)**2)/float(self.sigma0**2)

    def pdf(self,mu):
        if self.transform is not None:
            mu = self.transform(mu)             
        return (1/float(self.sigma0))*exp(-(0.5*(mu-self.mu0)**2)/float(self.sigma0**2))

class Uniform(object):

    def __init__(self,transform=None):
        self.transform_name = transform     
        self.transform = transform_define(transform)
        self.itransform = itransform_define(transform)

    def logpdf(self,mu):
        return 0.0

class InverseGamma(object):

    def __init__(self,alpha,beta,transform=np.exp):
        self.alpha = alpha
        self.beta = beta
        self.transform_name = transform
        self.transform = transform_define(transform)
        self.itransform = itransform_define(transform)

    def logpdf(self,x):
        if self.transform is not None:
            x = self.transform(x)       
        return (-self.alpha-1)*log(x) - (self.beta/float(x))

    def pdf(self,x):
        if self.transform is not None:
            x = self.transform(x)               
        return (x**(-self.alpha-1))*exp(-(self.beta/float(x)))





import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
import scipy.special as sp

from .. import inference as ifr
from .. import tsm as tsm
from .. import distributions as dst
from .. import data_check as dc

def exponential_link(x):
    return 1.0/np.exp(x)

class GASDistribution(object):

    def __init__(self,gradient_only=False):
        if gradient_only is True:
            self.score_function = self.first_order_score
        else:
            self.score_function = self.second_order_score


class GASExponential(GASDistribution):

    def __init__(self,gradient_only=True):
        if gradient_only is True:
            self.score_function = self.first_order_score
        else:
            self.score_function = self.second_order_score

    @staticmethod
    def setup():
        name = "Exponential GAS"
        link = exponential_link
        scale = False
        shape = False
        skewness = False
        mean_transform = np.log
        return name, link, scale, shape, skewness, mean_transform

    @staticmethod
    def build_parameters():
        parameters_to_build = []
        return parameters_to_build

    @staticmethod
    def draw_variable(loc,scale,shape,skewness,nsims):
        return np.random.exponential(1.0/loc, nsims)

    @staticmethod
    def first_order_score(y,mean,scale,shape,skewness):
        return 1 - (mean*y)

    @staticmethod
    def neg_loglikelihood(y,mean,scale,shape,skewness):
        return -np.sum(ss.expon.logpdf(x=y,scale=1/mean))

    @staticmethod
    def second_order_score(y,mean,scale,shape,skewness):
        return 1 - (mean*y)

    @staticmethod
    def reg_score_function(X,y,mean,scale,shape,skewness):
        return X*(1.0 - mean*y)


class GASLaplace(GASDistribution):

    def __init__(self,gradient_only=True):
        if gradient_only is True:
            self.score_function = self.first_order_score
        else:
            self.score_function = self.second_order_score

    @staticmethod
    def setup():
        name = "Laplace GAS"
        link = np.array
        scale = True
        shape = False
        skewness = False
        mean_transform = np.array
        return name, link, scale, shape, skewness, mean_transform

    @staticmethod
    def build_parameters():
        parameters_to_build = []
        parameters_to_build.append(['Laplace Scale',ifr.Uniform(transform='exp'),dst.q_Normal(0,3),2.0])
        return parameters_to_build

    @staticmethod
    def draw_variable(loc,scale,shape,skewness,nsims):
        return np.random.laplace(loc, scale, nsims)

    @staticmethod
    def first_order_score(y,mean,scale,shape,skewness):
        return (y-mean)/float(scale*np.abs(y-mean))

    @staticmethod
    def neg_loglikelihood(y,mean,scale,shape,skewness):
        return -np.sum(ss.laplace.logpdf(y,loc=mean,scale=scale))

    @staticmethod
    def second_order_score(y,mean,scale,shape,skewness):
        return ((y-mean)/float(scale*np.abs(y-mean))) / (-(np.power(y-mean,2) - np.power(np.abs(mean-y),2))/(scale*np.power(np.abs(mean-y),3)))

    @staticmethod
    def reg_score_function(X,y,mean,scale,shape,skewness):
        return X*(y-mean)/(scale*np.abs(y-mean))


class GASNormal(GASDistribution):

    @staticmethod
    def setup():
        name = "Normal GAS"
        link = np.array
        scale = True
        shape = False
        skewness = False
        mean_transform = np.array
        return name, link, scale, shape, skewness, mean_transform

    @staticmethod
    def build_parameters():
        parameters_to_build = []
        parameters_to_build.append(['Normal Scale',ifr.Uniform(transform='exp'),dst.q_Normal(0,3),0.0])
        return parameters_to_build

    @staticmethod
    def draw_variable(loc,scale,shape,skewness,nsims):
        return np.random.normal(loc, scale, nsims)

    @staticmethod
    def first_order_score(y,mean,scale,shape,skewness):
        return (y-mean)/np.power(scale,2)

    @staticmethod
    def neg_loglikelihood(y,mean,scale,shape,skewness):
        return -np.sum(ss.norm.logpdf(y,loc=mean,scale=scale))

    @staticmethod
    def second_order_score(y,mean,scale,shape,skewness):
        return y-mean

    @staticmethod
    def reg_score_function(X,y,mean,scale,shape,skewness):
        return X*(y-mean)


class GASPoisson(GASDistribution):

    @staticmethod
    def setup():
        name = "Poisson GAS"
        link = np.exp
        scale = False
        shape = False
        skewness = False
        mean_transform = np.log
        return name, link, scale, shape, skewness, mean_transform

    @staticmethod
    def build_parameters():
        parameters_to_build = []
        return parameters_to_build

    @staticmethod
    def draw_variable(loc,scale,shape,skewness,nsims):
        return np.random.poisson(loc, nsims)

    @staticmethod
    def first_order_score(y,mean,scale,shape,skewness):
        return y-mean

    @staticmethod
    def neg_loglikelihood(y,mean,scale,shape,skewness):
        return -np.sum(ss.poisson.logpmf(y,mean))

    @staticmethod
    def second_order_score(y,mean,scale,shape,skewness):
        return (y-mean)/float(mean)

    @staticmethod
    def reg_score_function(X,y,mean,scale,shape,skewness):
        return X*(y-mean)


class GASt(GASDistribution):

    def __init__(self,gradient_only=True):
        if gradient_only is True:
            self.score_function = self.first_order_score
        else:
            self.score_function = self.second_order_score

    @staticmethod
    def setup():
        name = "t GAS"
        link = np.array
        scale = True
        shape = True
        skewness = False
        mean_transform = np.array
        return name, link, scale, shape, skewness, mean_transform

    @staticmethod
    def build_parameters():
        parameters_to_build = []
        parameters_to_build.append(['t Scale',ifr.Uniform(transform='exp'),dst.q_Normal(0,3),0.0])
        parameters_to_build.append(['v',ifr.Uniform(transform='exp'),dst.q_Normal(0,3),2.0])
        return parameters_to_build

    @staticmethod
    def draw_variable(loc,scale,shape,skewness,nsims):
        return loc + scale*np.random.standard_t(shape,nsims)

    @staticmethod
    def first_order_score(y,mean,scale,shape,skewness):
        return ((shape+1)/shape)*(y-mean)/(np.power(scale,2) + (np.power(y-mean,2)/shape))

    @staticmethod
    def neg_loglikelihood(y,mean,scale,shape,skewness):
        return -np.sum(ss.t.logpdf(x=y,df=shape,loc=mean,scale=scale))

    @staticmethod
    def second_order_score(y,mean,scale,shape,skewness):
        return ((shape+1)/shape)*(y-mean)/(np.power(scale,2) + (np.power(y-mean,2)/shape))/((shape+1)*((np.power(scale,2)*shape) - np.power(y-mean,2))/np.power((np.power(scale,2)*shape) + np.power(y-mean,2),2))

    @staticmethod
    def reg_score_function(X,y,mean,scale,shape,skewness):
        return ((shape+1)/shape)*((y-mean)*X)/(np.power(scale,2)+np.power((y-mean),2)/shape)


class GASSkewt(GASDistribution):

    def __init__(self,gradient_only=True):
        if gradient_only is True:
            self.score_function = self.first_order_score
        else:
            self.score_function = self.second_order_score

    @staticmethod
    def setup():
        name = "Skewt GAS"
        link = np.array
        scale = True
        shape = True
        skewness = True
        mean_transform = np.array
        return name, link, scale, shape, skewness, mean_transform

    @staticmethod
    def build_parameters():
        parameters_to_build = []
        parameters_to_build.append(['Skewness',ifr.Uniform(transform='exp'),dst.q_Normal(0,3),0.0])
        parameters_to_build.append(['Skewt Scale',ifr.Uniform(transform='exp'),dst.q_Normal(0,3),0.0])
        parameters_to_build.append(['v',ifr.Uniform(transform='exp'),dst.q_Normal(0,3),2.0])
        return parameters_to_build

    @staticmethod
    def draw_variable(loc,scale,shape,skewness,nsims):
        return loc + scale*dst.skewt.rvs(shape,skewness,nsims)

    @staticmethod
    def first_order_score(y,mean,scale,shape,skewness):
        m1 = (np.sqrt(shape)*sp.gamma((shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(shape/2.0))
        mean = mean + (skewness - (1.0/skewness))*scale*m1
        if (y-mean)>=0:
            return ((shape+1)/shape)*(y-mean)/(np.power(skewness*scale,2) + (np.power(y-mean,2)/shape))
        else:
            return ((shape+1)/shape)*(y-mean)/(np.power(scale,2) + (np.power(skewness*(y-mean),2)/shape))

    @staticmethod
    def neg_loglikelihood(y,mean,scale,shape,skewness):
        m1 = (np.sqrt(shape)*sp.gamma((shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(shape/2.0))
        mean = mean + (skewness - (1.0/skewness))*scale*m1
        return -np.sum(dst.skewt.logpdf(x=y,df=shape,loc=mean,gamma=skewness,scale=scale))

    @staticmethod
    def second_order_score(y,mean,scale,shape,skewness):
        m1 = (np.sqrt(shape)*sp.gamma((shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(shape/2.0))
        mean = mean + (skewness - (1.0/skewness))*scale*m1
        if (y-mean)>=0:
            return ((shape+1)/shape)*(y-mean)/(np.power(skewness*scale,2) + (np.power(y-mean,2)/shape))
        else:
            return ((shape+1)/shape)*(y-mean)/(np.power(scale,2) + (np.power(skewness*(y-mean),2)/shape))

    @staticmethod
    def reg_score_function(X,y,mean,scale,shape,skewness):
        m1 = (np.sqrt(shape)*sp.gamma((shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(shape/2.0))
        mean = mean + (skewness - (1.0/skewness))*scale*m1
        if (y-mean)>=0:
            return ((shape+1)/shape)*((y-mean)*X)/(np.power(skewness*scale,2) + (np.power(y-mean,2)/shape))
        else:
            return ((shape+1)/shape)*((y-mean)*X)/(np.power(scale,2) + (np.power(skewness*(y-mean),2)/shape))





import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
from scipy.stats import multivariate_normal
import matplotlib.pyplot as plt
import seaborn as sns

from .. import arma
from .. import inference as ifr
from .. import distributions as dst
from .. import output as op
from .. import tests as tst
from .. import tsm as tsm
from .. import data_check as dc

from .kernels import *

class GP(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** GAUSSIAN PROCESS MODELS ****
    """

    def __init__(self):

        # Initialize TSM object
        super(GP,self).__init__('GP')

    def _alpha(self,L):
        """ Covariance-derived term to construct expectations. See Rasmussen & Williams.

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        The alpha matrix/vector
        """     

        return np.linalg.solve(np.transpose(L),np.linalg.solve(L,np.transpose(self.data)))

    def _L(self,beta):
        """ Creates cholesky decomposition of covariance matrix

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        The cholesky decomposition (L) of K
        """ 

        return np.linalg.cholesky(self.kernel.K()) + np.identity(self.kernel.K().shape[0])*self.parameters.parameter_list[0].prior.transform(beta[0])

    def expected_values(self,beta):
        """ Expected values of the function given the covariance matrix and hyperparameters

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        The expected values of the function
        """     

        self._start_params(beta)
        L = self._L(beta)
        return np.dot(np.transpose(self.kernel.K()),self._alpha(L))

    def variance_values(self,beta):
        """ Covariance matrix for the estimated function

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        Covariance matrix for the estimated function 
        """     

        self._start_params(beta)

        return self.kernel.K() - np.dot(np.dot(np.transpose(self.kernel.K()),np.linalg.pinv(self.kernel.K() + np.identity(self.kernel.K().shape[0])*self.parameters.parameter_list[0].prior.transform(beta[0]))),self.kernel.K()) + self.parameters.parameter_list[0].prior.transform(beta[0])

    def neg_loglik(self,beta):
        """ Creates the negative log marginal likelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for parameters

        Returns
        ----------
        The negative log marginal logliklihood of the model
        """             
        self._start_params(beta)
        L = self._L(beta)
        return -(-0.5*(np.dot(np.transpose(self.data),self._alpha(L))) - np.trace(L) - (self.data.shape[0]/2)*np.log(2*np.pi))

    def plot_predict(self,h=5,past_values=20,intervals=True,**kwargs):

        """ Plots forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show 95% prediction intervals for the forecast?

        Returns
        ----------
        - Plot of the forecast
        - Error bars, forecasted_values, plot_values, plot_index
        """     

        figsize = kwargs.get('figsize',(10,7))

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:

            predictions, variance, lower, upper = self._construct_predict(self.parameters.get_parameter_values(),h) 
            full_predictions = np.append(self.data,predictions)
            full_lower = np.append(self.data,lower)
            full_upper = np.append(self.data,upper)
            date_index = self.shift_dates(h)

            # Plot values (how far to look back)
            plot_values = full_predictions[-h-past_values:]*self._norm_std + self._norm_mean
            plot_index = date_index[-h-past_values:]

            # Lower and upper intervals
            lower = np.append(full_predictions[-h-1],lower)
            upper = np.append(full_predictions[-h-1],upper)

            plt.figure(figsize=figsize)
            if intervals == True:
                plt.fill_between(date_index[-h-1:], 
                    lower*self._norm_std + self._norm_mean, 
                    upper*self._norm_std + self._norm_mean,
                    alpha=0.2)          
            
            plt.plot(plot_index,plot_values)
            plt.title("Forecast for " + self.data_name)
            plt.xlabel("Time")
            plt.ylabel(self.data_name)
            plt.show()

    def plot_predict_is(self,h=5,**kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        Returns
        ----------
        - Plot of the forecast against data 
        """     

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        date_index = self.index[-h:]
        predictions = self.predict_is(h)
        data = self.data[-h:]

        plt.plot(date_index,data*self._norm_std + self._norm_mean,label='Data')
        plt.plot(date_index,predictions,label='Predictions',c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def predict(self,h=5):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        if self.parameters.estimated is False:
            raise Exception("No parameters estimated!")
        else:

            predictions, _, _, _ = self._construct_predict(self.parameters.get_parameter_values(),h)    
            predictions = predictions*self._norm_std + self._norm_mean  
            date_index = self.shift_dates(h)
            result = pd.DataFrame(predictions)
            result.rename(columns={0:self.data_name}, inplace=True)
            result.index = date_index[-h:]

            return result


